from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

class SDL_AsyncIO(ctypes.c_void_p):
    ...

SDL_AsyncIOTaskType: typing.TypeAlias = SDL_TYPE["SDL_AsyncIOTaskType", SDL_ENUM]

SDL_ASYNCIO_TASK_READ, SDL_ASYNCIO_TASK_WRITE, SDL_ASYNCIO_TASK_CLOSE = range(3)

SDL_AsyncIOResult: typing.TypeAlias = SDL_TYPE["SDL_AsyncIOResult", SDL_ENUM]

SDL_ASYNCIO_COMPLETE, SDL_ASYNCIO_FAILURE, SDL_ASYNCIO_CANCELED = range(3)

class SDL_AsyncIOOutcome(ctypes.Structure):
    _fields_ = [
        ("asyncio", SDL_POINTER[SDL_AsyncIO]),
        ("type", SDL_AsyncIOTaskType),
        ("result", SDL_AsyncIOResult),
        ("buffer", ctypes.c_void_p),
        ("offset", ctypes.c_uint64),
        ("bytes_requested", ctypes.c_uint64),
        ("bytes_transferred", ctypes.c_uint64),
        ("userdata", ctypes.c_void_p)
    ]

class SDL_AsyncIOQueue(ctypes.c_void_p):
    ...

SDL_AsyncIOFromFile: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AsyncIOFromFile", SDL_POINTER[SDL_AsyncIO], [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_GetAsyncIOSize: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAsyncIOSize", ctypes.c_int64, [SDL_POINTER[SDL_AsyncIO]], SDL_BINARY]

SDL_ReadAsyncIO: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ReadAsyncIO", ctypes.c_bool, [SDL_POINTER[SDL_AsyncIO], ctypes.c_void_p, ctypes.c_uint64, ctypes.c_uint64, SDL_POINTER[SDL_AsyncIOQueue], ctypes.c_void_p], SDL_BINARY]
SDL_WriteAsyncIO: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_WriteAsyncIO", ctypes.c_bool, [SDL_POINTER[SDL_AsyncIO], ctypes.c_void_p, ctypes.c_uint64, ctypes.c_uint64, SDL_POINTER[SDL_AsyncIOQueue], ctypes.c_void_p], SDL_BINARY]
SDL_CloseAsyncIO: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CloseAsyncIO", ctypes.c_bool, [SDL_POINTER[SDL_AsyncIO], ctypes.c_bool, SDL_POINTER[SDL_AsyncIOQueue], ctypes.c_void_p], SDL_BINARY]

SDL_CreateAsyncIOQueue: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateAsyncIOQueue", SDL_POINTER[SDL_AsyncIOQueue], [], SDL_BINARY]
SDL_DestroyAsyncIOQueue: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DestroyAsyncIOQueue", None, [SDL_POINTER[SDL_AsyncIOQueue]], SDL_BINARY]
SDL_GetAsyncIOResult: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAsyncIOResult", ctypes.c_bool, [SDL_POINTER[SDL_AsyncIOQueue], SDL_POINTER[SDL_AsyncIOOutcome]], SDL_BINARY]
SDL_WaitAsyncIOResult: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_WaitAsyncIOResult", ctypes.c_bool, [SDL_POINTER[SDL_AsyncIOQueue], SDL_POINTER[SDL_AsyncIOOutcome], ctypes.c_int32], SDL_BINARY]
SDL_SignalAsyncIOQueue: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SignalAsyncIOQueue", None, [SDL_POINTER[SDL_AsyncIOQueue]], SDL_BINARY]

SDL_LoadFileAsync: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_LoadFileAsync", ctypes.c_bool, [ctypes.c_char_p, SDL_POINTER[SDL_AsyncIOQueue], ctypes.c_void_p], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_FUNC_TYPE, SDL_FUNC, SDL_BINARY

SDL_SetClipboardText: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetClipboardText", ctypes.c_bool, [ctypes.c_char_p], SDL_BINARY]
SDL_GetClipboardText: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetClipboardText", ctypes.c_char_p, [], SDL_BINARY]
SDL_HasClipboardText: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_HasClipboardText", ctypes.c_bool, [], SDL_BINARY]

SDL_SetPrimarySelectionText: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetPrimarySelectionText", ctypes.c_bool, [ctypes.c_char_p], SDL_BINARY]
SDL_GetPrimarySelectionText: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPrimarySelectionText", ctypes.c_char_p, [], SDL_BINARY]
SDL_HasPrimarySelectionText: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_HasPrimarySelectionText", ctypes.c_bool, [], SDL_BINARY]

SDL_ClipboardDataCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_ClipboardDataCallback", ctypes.c_void_p, [ctypes.c_char_p, SDL_POINTER[ctypes.c_size_t]]]
SDL_ClipboardCleanupCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_ClipboardCleanupCallback", None, [ctypes.c_void_p]]

SDL_SetClipboardData: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetClipboardData", ctypes.c_bool, [SDL_ClipboardDataCallback, SDL_ClipboardCleanupCallback, ctypes.c_void_p, SDL_POINTER[ctypes.c_char_p], ctypes.c_size_t], SDL_BINARY]
SDL_ClearClipboardData: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ClearClipboardData", ctypes.c_bool, [], SDL_BINARY]
SDL_GetClipboardData: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetClipboardData", ctypes.c_void_p, [ctypes.c_char_p, SDL_POINTER[ctypes.c_size_t]], SDL_BINARY]
SDL_HasClipboardData: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_HasClipboardData", ctypes.c_bool, [ctypes.c_char_p], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_FUNC_TYPE, SDL_FUNC, SDL_TYPE, SDL_BINARY

from .SDL_surface import SDL_Surface

class SDL_Tray(ctypes.c_void_p):
    ...

class SDL_TrayMenu(ctypes.c_void_p):
    ...

class SDL_TrayEntry(ctypes.c_void_p):
    ...

SDL_TrayEntryFlags: typing.TypeAlias = SDL_TYPE["SDL_TrayEntryFlags", ctypes.c_uint32]

SDL_TRAYENTRY_BUTTON, SDL_TRAYENTRY_CHECKBOX, SDL_TRAYENTRY_SUBMENU, SDL_TRAYENTRY_DISABLED, \
    SDL_TRAYENTRY_CHECKED = 0x00000001, 0x00000002, 0x00000004, 0x80000000, 0x40000000

SDL_TrayCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_TrayCallback", None, [ctypes.c_void_p, SDL_POINTER[SDL_TrayEntry]]]

SDL_CreateTray: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateTray", SDL_POINTER[SDL_Tray], [SDL_POINTER[SDL_Surface], ctypes.c_char_p], SDL_BINARY]
SDL_SetTrayIcon: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetTrayIcon", None, [SDL_POINTER[SDL_Tray], SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_SetTrayTooltip: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetTrayTooltip", None, [SDL_POINTER[SDL_Tray], ctypes.c_char_p], SDL_BINARY]
SDL_CreateTrayMenu: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateTrayMenu", SDL_POINTER[SDL_TrayMenu], [SDL_POINTER[SDL_Tray]], SDL_BINARY]
SDL_CreateTraySubmenu: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateTraySubmenu", SDL_POINTER[SDL_TrayMenu], [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_GetTrayMenu: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayMenu", SDL_POINTER[SDL_TrayMenu], [SDL_POINTER[SDL_Tray]], SDL_BINARY]
SDL_GetTraySubmenu: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTraySubmenu", SDL_POINTER[SDL_TrayMenu], [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_GetTrayEntries: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayEntries", SDL_POINTER[SDL_POINTER[SDL_TrayEntry]], [SDL_POINTER[SDL_TrayMenu], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_RemoveTrayEntry: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RemoveTrayEntry", None, [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_InsertTrayEntryAt: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_InsertTrayEntryAt", SDL_POINTER[SDL_TrayEntry], [SDL_POINTER[SDL_TrayMenu], ctypes.c_int, ctypes.c_char_p, SDL_TrayEntryFlags], SDL_BINARY]
SDL_SetTrayEntryLabel: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetTrayEntryLabel", None, [SDL_POINTER[SDL_TrayEntry], ctypes.c_char_p], SDL_BINARY]
SDL_GetTrayEntryLabel: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayEntryLabel", ctypes.c_char_p, [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_SetTrayEntryChecked: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetTrayEntryChecked", None, [SDL_POINTER[SDL_TrayEntry], ctypes.c_bool], SDL_BINARY]
SDL_GetTrayEntryChecked: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayEntryChecked", ctypes.c_bool, [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_SetTrayEntryEnabled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetTrayEntryEnabled", None, [SDL_POINTER[SDL_TrayEntry], ctypes.c_bool], SDL_BINARY]
SDL_GetTrayEntryEnabled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayEntryEnabled", ctypes.c_bool, [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_SetTrayEntryCallback: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetTrayEntryCallback", None, [SDL_POINTER[SDL_TrayEntry], SDL_TrayCallback, ctypes.c_void_p], SDL_BINARY]
SDL_ClickTrayEntry: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ClickTrayEntry", None, [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_DestroyTray: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DestroyTray", None, [SDL_POINTER[SDL_Tray]], SDL_BINARY]
SDL_GetTrayEntryParent: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayEntryParent", SDL_POINTER[SDL_TrayMenu], [SDL_POINTER[SDL_TrayEntry]], SDL_BINARY]
SDL_GetTrayMenuParentEntry: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayMenuParentEntry", SDL_POINTER[SDL_TrayEntry], [SDL_POINTER[SDL_TrayMenu]], SDL_BINARY]
SDL_GetTrayMenuParentTray: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTrayMenuParentTray", SDL_POINTER[SDL_Tray], [SDL_POINTER[SDL_TrayMenu]], SDL_BINARY]
SDL_UpdateTrays: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UpdateTrays", None, [], SDL_BINARY]


from .__init__ import ctypes, typing, abc, SDL_FUNC_TYPE, \
    SDL_POINTER, SDL_FUNC, SDL_TYPE, SDL_BINARY, SDL_ENUM

from .SDL_properties import SDL_PropertiesID
from .SDL_endian import SDL_BYTEORDER, SDL_LIL_ENDIAN
from .SDL_iostream import SDL_IOStream

SDL_AUDIO_MASK_BITSIZE: int = 0xFF
SDL_AUDIO_MASK_FLOAT: int = 1 << 8
SDL_AUDIO_MASK_BIG_ENDIAN: int = 1 << 12
SDL_AUDIO_MASK_SIGNED: int = 1 << 15

SDL_DEFINE_AUDIO_FORMAT: abc.Callable[[int, int, int, int], int] = lambda signed, bigendian, _float, size: \
    (signed << 15) | (bigendian << 12) | (_float << 8) | (size & SDL_AUDIO_MASK_BITSIZE)

SDL_AudioFormat: typing.TypeAlias = SDL_TYPE["SDL_AudioFormat", SDL_ENUM]

SDL_AUDIO_UNKNOWN: int = 0x0000
SDL_AUDIO_U8: int = 0x0008
SDL_AUDIO_S8: int = 0x8008
SDL_AUDIO_S16LE: int = 0x8010
SDL_AUDIO_S16BE: int = 0x9010
SDL_AUDIO_S32LE: int = 0x8020
SDL_AUDIO_S32BE: int = 0x9020
SDL_AUDIO_F32LE: int = 0x8120
SDL_AUDIO_F32BE: int = 0x9120

if SDL_BYTEORDER == SDL_LIL_ENDIAN:
    SDL_AUDIO_S16, SDL_AUDIO_S32, SDL_AUDIO_F32 = SDL_AUDIO_S16LE, SDL_AUDIO_S32LE, SDL_AUDIO_F32LE

else:
    SDL_AUDIO_S16, SDL_AUDIO_S32, SDL_AUDIO_F32 = SDL_AUDIO_S16BE, SDL_AUDIO_S32BE, SDL_AUDIO_F32BE

SDL_AUDIO_BITSIZE: abc.Callable[[SDL_AudioFormat], int] = lambda x: x & SDL_AUDIO_MASK_BITSIZE
SDL_AUDIO_BYTESIZE: abc.Callable[[SDL_AudioFormat], int] = lambda x: SDL_AUDIO_BITSIZE(x) / 8

SDL_AUDIO_ISFLOAT: abc.Callable[[SDL_AudioFormat], int] = lambda x: x & SDL_AUDIO_MASK_FLOAT
SDL_AUDIO_ISBIGENDIAN: abc.Callable[[SDL_AudioFormat], int] = lambda x: x & SDL_AUDIO_MASK_BIG_ENDIAN
SDL_AUDIO_ISLITTLEENDIAN: abc.Callable[[SDL_AudioFormat], bool] = lambda x: not SDL_AUDIO_ISBIGENDIAN(x)
SDL_AUDIO_ISSIGNED: abc.Callable[[SDL_AudioFormat], int] = lambda x: x & SDL_AUDIO_MASK_SIGNED
SDL_AUDIO_ISINT: abc.Callable[[SDL_AudioFormat], bool] = lambda x: not SDL_AUDIO_ISFLOAT(x)
SDL_AUDIO_ISUNSIGNED: abc.Callable[[SDL_AudioFormat], bool] = lambda x: not SDL_AUDIO_ISSIGNED(x)

SDL_AudioDeviceID: typing.TypeAlias = SDL_TYPE["SDL_AudioDeviceID", ctypes.c_uint32]

SDL_AUDIO_DEVICE_DEFAULT_PLAYBACK: int = 0xFFFFFFFF
SDL_AUDIO_DEVICE_DEFAULT_RECORDING: int = 0xFFFFFFFE

class SDL_AudioSpec(ctypes.Structure):
    _fields_ = [
        ("format", SDL_AudioFormat),
        ("channels", ctypes.c_int),
        ("freq", ctypes.c_int)
    ]

SDL_AUDIO_FRAMESIZE: abc.Callable[[SDL_AudioSpec], int] = lambda x: SDL_AUDIO_BYTESIZE(x.format) * x.channels

class SDL_AudioStream(ctypes.c_void_p):
    ...

SDL_GetNumAudioDrivers: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetNumAudioDrivers", ctypes.c_int, [], SDL_BINARY]
SDL_GetAudioDriver: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioDriver", ctypes.c_char_p, [ctypes.c_int], SDL_BINARY]
SDL_GetCurrentAudioDriver: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetCurrentAudioDriver", ctypes.c_char_p, [], SDL_BINARY]

SDL_GetAudioPlaybackDevices: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioPlaybackDevices", SDL_POINTER[SDL_AudioDeviceID], [SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_GetAudioRecordingDevices: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioRecordingDevices", SDL_POINTER[SDL_AudioDeviceID], [SDL_POINTER[ctypes.c_int]], SDL_BINARY]

SDL_GetAudioDeviceName: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioDeviceName", ctypes.c_char_p, [SDL_AudioDeviceID], SDL_BINARY]
SDL_GetAudioDeviceFormat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioDeviceFormat", ctypes.c_bool, [SDL_AudioDeviceID, SDL_POINTER[SDL_AudioSpec], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_GetAudioDeviceChannelMap: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioDeviceChannelMap", SDL_POINTER[ctypes.c_int], [SDL_AudioDeviceID, SDL_POINTER[ctypes.c_int]], SDL_BINARY]

SDL_OpenAudioDevice: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OpenAudioDevice", SDL_AudioDeviceID, [SDL_AudioDeviceID, SDL_POINTER[SDL_AudioSpec]], SDL_BINARY]
SDL_IsAudioDevicePhysical: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_IsAudioDevicePhysical", ctypes.c_bool, [SDL_AudioDeviceID], SDL_BINARY]
SDL_IsAudioDevicePlayback: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_IsAudioDevicePlayback", ctypes.c_bool, [SDL_AudioDeviceID], SDL_BINARY]
SDL_PauseAudioDevice: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_PauseAudioDevice", ctypes.c_bool, [SDL_AudioDeviceID], SDL_BINARY]
SDL_ResumeAudioDevice: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ResumeAudioDevice", ctypes.c_bool, [SDL_AudioDeviceID], SDL_BINARY]
SDL_AudioDevicePaused: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AudioDevicePaused", ctypes.c_bool, [SDL_AudioDeviceID], SDL_BINARY]
SDL_GetAudioDeviceGain: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioDeviceGain", ctypes.c_float, [SDL_AudioDeviceID], SDL_BINARY]
SDL_SetAudioDeviceGain: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioDeviceGain", ctypes.c_bool, [SDL_AudioDeviceID, ctypes.c_float], SDL_BINARY]
SDL_CloseAudioDevice: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CloseAudioDevice", None, [SDL_AudioDeviceID], SDL_BINARY]

SDL_BindAudioStreams: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BindAudioStreams", ctypes.c_bool, [SDL_AudioDeviceID, SDL_POINTER[SDL_POINTER[SDL_AudioStream]], ctypes.c_int], SDL_BINARY]
SDL_BindAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BindAudioStream", ctypes.c_bool, [SDL_AudioDeviceID, SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_UnbindAudioStreams: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UnbindAudioStreams", None, [SDL_POINTER[SDL_POINTER[SDL_AudioStream]], ctypes.c_int], SDL_BINARY]
SDL_UnbindAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UnbindAudioStream", None, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_GetAudioStreamDevice: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamDevice", SDL_AudioDeviceID, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_CreateAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateAudioStream", SDL_POINTER[SDL_AudioStream], [SDL_POINTER[SDL_AudioSpec], SDL_POINTER[SDL_AudioSpec]], SDL_BINARY]
SDL_GetAudioStreamProperties: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamProperties", SDL_PropertiesID, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_GetAudioStreamFormat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamFormat", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], SDL_POINTER[SDL_AudioSpec], SDL_POINTER[SDL_AudioSpec]], SDL_BINARY]
SDL_SetAudioStreamFormat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioStreamFormat", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], SDL_POINTER[SDL_AudioSpec], SDL_POINTER[SDL_AudioSpec]], SDL_BINARY]
SDL_GetAudioStreamFrequencyRatio: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamFrequencyRatio", ctypes.c_float, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_SetAudioStreamFrequencyRatio: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioStreamFrequencyRatio", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], ctypes.c_float], SDL_BINARY]
SDL_GetAudioStreamGain: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamGain", ctypes.c_float, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_SetAudioStreamGain: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioStreamGain", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], ctypes.c_float], SDL_BINARY]
SDL_GetAudioStreamInputChannelMap: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamInputChannelMap", SDL_POINTER[ctypes.c_int], [SDL_POINTER[SDL_AudioStream], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_GetAudioStreamOutputChannelMap: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamOutputChannelMap", SDL_POINTER[ctypes.c_int], [SDL_POINTER[SDL_AudioStream], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_SetAudioStreamInputChannelMap: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioStreamInputChannelMap", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], SDL_POINTER[ctypes.c_int], ctypes.c_int], SDL_BINARY]
SDL_SetAudioStreamOutputChannelMap: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioStreamOutputChannelMap", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], SDL_POINTER[ctypes.c_int], ctypes.c_int], SDL_BINARY]
SDL_PutAudioStreamData: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_PutAudioStreamData", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], ctypes.c_void_p, ctypes.c_int], SDL_BINARY]
SDL_GetAudioStreamData: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamData", ctypes.c_int, [SDL_POINTER[SDL_AudioStream], ctypes.c_void_p, ctypes.c_int], SDL_BINARY]
SDL_GetAudioStreamAvailable: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamAvailable", ctypes.c_int, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_GetAudioStreamQueued: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioStreamQueued", ctypes.c_int, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_FlushAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_FlushAudioStream", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_ClearAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ClearAudioStream", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_PauseAudioStreamDevice: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_PauseAudioStreamDevice", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_ResumeAudioStreamDevice: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ResumeAudioStreamDevice", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_AudioStreamDevicePaused: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AudioStreamDevicePaused", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_LockAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_LockAudioStream", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_UnlockAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UnlockAudioStream", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]

SDL_AudioStreamCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_AudioStreamCallback", None, [ctypes.c_void_p, SDL_POINTER[SDL_AudioStream], ctypes.c_int, ctypes.c_int]]

SDL_SetAudioStreamGetCallback: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioStreamGetCallback", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], SDL_AudioStreamCallback, ctypes.c_void_p], SDL_BINARY]
SDL_SetAudioStreamPutCallback: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioStreamPutCallback", ctypes.c_bool, [SDL_POINTER[SDL_AudioStream], SDL_AudioStreamCallback, ctypes.c_void_p], SDL_BINARY]
SDL_DestroyAudioStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DestroyAudioStream", None, [SDL_POINTER[SDL_AudioStream]], SDL_BINARY]
SDL_OpenAudioDeviceStream: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OpenAudioDeviceStream", SDL_POINTER[SDL_AudioStream], [SDL_AudioDeviceID, SDL_POINTER[SDL_AudioSpec], SDL_AudioStreamCallback, ctypes.c_void_p], SDL_BINARY]

SDL_AudioPostmixCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_AudioPostmixCallback", None, [ctypes.c_void_p, SDL_POINTER[SDL_AudioSpec], SDL_POINTER[ctypes.c_float], ctypes.c_int]]

SDL_SetAudioPostmixCallback: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAudioPostmixCallback", ctypes.c_bool, [SDL_AudioDeviceID, SDL_AudioPostmixCallback, ctypes.c_void_p], SDL_BINARY]

SDL_LoadWAV_IO: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_LoadWAV_IO", ctypes.c_bool, [SDL_POINTER[SDL_IOStream], ctypes.c_bool, SDL_POINTER[SDL_AudioSpec], SDL_POINTER[SDL_POINTER[ctypes.c_uint8]], SDL_POINTER[ctypes.c_uint32]], SDL_BINARY]
SDL_LoadWAV: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_LoadWAV", ctypes.c_bool, [ctypes.c_char_p, SDL_POINTER[SDL_AudioSpec], SDL_POINTER[SDL_POINTER[ctypes.c_uint8]], SDL_POINTER[ctypes.c_uint32]], SDL_BINARY]
SDL_MixAudio: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_MixAudio", ctypes.c_bool, [SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8], SDL_AudioFormat, ctypes.c_uint32, ctypes.c_float], SDL_BINARY]
SDL_ConvertAudioSamples: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ConvertAudioSamples", ctypes.c_bool, [SDL_POINTER[SDL_AudioSpec], SDL_POINTER[ctypes.c_uint8], ctypes.c_int, SDL_POINTER[SDL_AudioSpec], SDL_POINTER[SDL_POINTER[ctypes.c_uint8]], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_GetAudioFormatName: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAudioFormatName", ctypes.c_char_p, [SDL_AudioFormat], SDL_BINARY]
SDL_GetSilenceValueForFormat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSilenceValueForFormat", ctypes.c_int, [SDL_AudioFormat], SDL_BINARY]


from .__init__ import ctypes, typing, abc, SDL_TYPE

from .SDL_scancode import SDL_Scancode

SDL_Keycode: typing.TypeAlias = SDL_TYPE["SDL_Keycode", ctypes.c_uint32]

SDLK_EXTENDED_MASK, SDLK_SCANCODE_MASK = 1 << 29, 1 << 30
SDL_SCANCODE_TO_KEYCODE: abc.Callable[[SDL_Scancode], SDL_Keycode] = lambda x: x | SDLK_SCANCODE_MASK

SDLK_UNKNOWN: int = 0x00000000
SDLK_RETURN: int = 0x0000000d
SDLK_ESCAPE: int = 0x0000001b
SDLK_BACKSPACE: int = 0x00000008
SDLK_TAB: int = 0x00000009
SDLK_SPACE: int = 0x00000020
SDLK_EXCLAIM: int = 0x00000021
SDLK_DBLAPOSTROPHE: int = 0x00000022
SDLK_HASH: int = 0x00000023
SDLK_DOLLAR: int = 0x00000024
SDLK_PERCENT: int = 0x00000025
SDLK_AMPERSAND: int = 0x00000026
SDLK_APOSTROPHE: int = 0x00000027
SDLK_LEFTPAREN: int = 0x00000028
SDLK_RIGHTPAREN: int = 0x00000029
SDLK_ASTERISK: int = 0x0000002a
SDLK_PLUS: int = 0x0000002b
SDLK_COMMA: int = 0x0000002c
SDLK_MINUS: int = 0x0000002d
SDLK_PERIOD: int = 0x0000002e
SDLK_SLASH: int = 0x0000002f
SDLK_0: int = 0x00000030
SDLK_1: int = 0x00000031
SDLK_2: int = 0x00000032
SDLK_3: int = 0x00000033
SDLK_4: int = 0x00000034
SDLK_5: int = 0x00000035
SDLK_6: int = 0x00000036
SDLK_7: int = 0x00000037
SDLK_8: int = 0x00000038
SDLK_9: int = 0x00000039
SDLK_COLON: int = 0x0000003a
SDLK_SEMICOLON: int = 0x0000003b
SDLK_LESS: int = 0x0000003c
SDLK_EQUALS: int = 0x0000003d
SDLK_GREATER: int = 0x0000003e
SDLK_QUESTION: int = 0x0000003f
SDLK_AT: int = 0x00000040
SDLK_LEFTBRACKET: int = 0x0000005b
SDLK_BACKSLASH: int = 0x0000005c
SDLK_RIGHTBRACKET: int = 0x0000005d
SDLK_CARET: int = 0x0000005e
SDLK_UNDERSCORE: int = 0x0000005f
SDLK_GRAVE: int = 0x00000060
SDLK_A: int = 0x00000061
SDLK_B: int = 0x00000062
SDLK_C: int = 0x00000063
SDLK_D: int = 0x00000064
SDLK_E: int = 0x00000065
SDLK_F: int = 0x00000066
SDLK_G: int = 0x00000067
SDLK_H: int = 0x00000068
SDLK_I: int = 0x00000069
SDLK_J: int = 0x0000006a
SDLK_K: int = 0x0000006b
SDLK_L: int = 0x0000006c
SDLK_M: int = 0x0000006d
SDLK_N: int = 0x0000006e
SDLK_O: int = 0x0000006f
SDLK_P: int = 0x00000070
SDLK_Q: int = 0x00000071
SDLK_R: int = 0x00000072
SDLK_S: int = 0x00000073
SDLK_T: int = 0x00000074
SDLK_U: int = 0x00000075
SDLK_V: int = 0x00000076
SDLK_W: int = 0x00000077
SDLK_X: int = 0x00000078
SDLK_Y: int = 0x00000079
SDLK_Z: int = 0x0000007a
SDLK_LEFTBRACE: int = 0x0000007b
SDLK_PIPE: int = 0x0000007c
SDLK_RIGHTBRACE: int = 0x0000007d
SDLK_TILDE: int = 0x0000007e
SDLK_DELETE: int = 0x0000007f
SDLK_PLUSMINUS: int = 0x000000b1
SDLK_CAPSLOCK: int = 0x40000039
SDLK_F1: int = 0x4000003a
SDLK_F2: int = 0x4000003b
SDLK_F3: int = 0x4000003c
SDLK_F4: int = 0x4000003d
SDLK_F5: int = 0x4000003e
SDLK_F6: int = 0x4000003f
SDLK_F7: int = 0x40000040
SDLK_F8: int = 0x40000041
SDLK_F9: int = 0x40000042
SDLK_F10: int = 0x40000043
SDLK_F11: int = 0x40000044
SDLK_F12: int = 0x40000045
SDLK_PRINTSCREEN: int = 0x40000046
SDLK_SCROLLLOCK: int = 0x40000047
SDLK_PAUSE: int = 0x40000048
SDLK_INSERT: int = 0x40000049
SDLK_HOME: int = 0x4000004a
SDLK_PAGEUP: int = 0x4000004b
SDLK_END: int = 0x4000004d
SDLK_PAGEDOWN: int = 0x4000004e
SDLK_RIGHT: int = 0x4000004f
SDLK_LEFT: int = 0x40000050
SDLK_DOWN: int = 0x40000051
SDLK_UP: int = 0x40000052
SDLK_NUMLOCKCLEAR: int = 0x40000053
SDLK_KP_DIVIDE: int = 0x40000054
SDLK_KP_MULTIPLY: int = 0x40000055
SDLK_KP_MINUS: int = 0x40000056
SDLK_KP_PLUS: int = 0x40000057
SDLK_KP_ENTER: int = 0x40000058
SDLK_KP_1: int = 0x40000059
SDLK_KP_2: int = 0x4000005a
SDLK_KP_3: int = 0x4000005b
SDLK_KP_4: int = 0x4000005c
SDLK_KP_5: int = 0x4000005d
SDLK_KP_6: int = 0x4000005e
SDLK_KP_7: int = 0x4000005f
SDLK_KP_8: int = 0x40000060
SDLK_KP_9: int = 0x40000061
SDLK_KP_0: int = 0x40000062
SDLK_KP_PERIOD: int = 0x40000063
SDLK_APPLICATION: int = 0x40000065
SDLK_POWER: int = 0x40000066
SDLK_KP_EQUALS: int = 0x40000067
SDLK_F13: int = 0x40000068
SDLK_F14: int = 0x40000069
SDLK_F15: int = 0x4000006a
SDLK_F16: int = 0x4000006b
SDLK_F17: int = 0x4000006c
SDLK_F18: int = 0x4000006d
SDLK_F19: int = 0x4000006e
SDLK_F20: int = 0x4000006f
SDLK_F21: int = 0x40000070
SDLK_F22: int = 0x40000071
SDLK_F23: int = 0x40000072
SDLK_F24: int = 0x40000073
SDLK_EXECUTE: int = 0x40000074
SDLK_HELP: int = 0x40000075
SDLK_MENU: int = 0x40000076
SDLK_SELECT: int = 0x40000077
SDLK_STOP: int = 0x40000078
SDLK_AGAIN: int = 0x40000079
SDLK_UNDO: int = 0x4000007a
SDLK_CUT: int = 0x4000007b
SDLK_COPY: int = 0x4000007c
SDLK_PASTE: int = 0x4000007d
SDLK_FIND: int = 0x4000007e
SDLK_MUTE: int = 0x4000007f
SDLK_VOLUMEUP: int = 0x40000080
SDLK_VOLUMEDOWN: int = 0x40000081
SDLK_KP_COMMA: int = 0x40000085
SDLK_KP_EQUALSAS400: int = 0x40000086
SDLK_ALTERASE: int = 0x40000099
SDLK_SYSREQ: int = 0x4000009a
SDLK_CANCEL: int = 0x4000009b
SDLK_CLEAR: int = 0x4000009c
SDLK_PRIOR: int = 0x4000009d
SDLK_RETURN2: int = 0x4000009e
SDLK_SEPARATOR: int = 0x4000009f
SDLK_OUT: int = 0x400000a0
SDLK_OPER: int = 0x400000a1
SDLK_CLEARAGAIN: int = 0x400000a2
SDLK_CRSEL: int = 0x400000a3
SDLK_EXSEL: int = 0x400000a4
SDLK_KP_00: int = 0x400000b0
SDLK_KP_000: int = 0x400000b1
SDLK_THOUSANDSSEPARATOR: int = 0x400000b2
SDLK_DECIMALSEPARATOR: int = 0x400000b3
SDLK_CURRENCYUNIT: int = 0x400000b4
SDLK_CURRENCYSUBUNIT: int = 0x400000b5
SDLK_KP_LEFTPAREN: int = 0x400000b6
SDLK_KP_RIGHTPAREN: int = 0x400000b7
SDLK_KP_LEFTBRACE: int = 0x400000b8
SDLK_KP_RIGHTBRACE: int = 0x400000b9
SDLK_KP_TAB: int = 0x400000ba
SDLK_KP_BACKSPACE: int = 0x400000bb
SDLK_KP_A: int = 0x400000bc
SDLK_KP_B: int = 0x400000bd
SDLK_KP_C: int = 0x400000be
SDLK_KP_D: int = 0x400000bf
SDLK_KP_E: int = 0x400000c0
SDLK_KP_F: int = 0x400000c1
SDLK_KP_XOR: int = 0x400000c2
SDLK_KP_POWER: int = 0x400000c3
SDLK_KP_PERCENT: int = 0x400000c4
SDLK_KP_LESS: int = 0x400000c5
SDLK_KP_GREATER: int = 0x400000c6
SDLK_KP_AMPERSAND: int = 0x400000c7
SDLK_KP_DBLAMPERSAND: int = 0x400000c8
SDLK_KP_VERTICALBAR: int = 0x400000c9
SDLK_KP_DBLVERTICALBAR: int = 0x400000ca
SDLK_KP_COLON: int = 0x400000cb
SDLK_KP_HASH: int = 0x400000cc
SDLK_KP_SPACE: int = 0x400000cd
SDLK_KP_AT: int = 0x400000ce
SDLK_KP_EXCLAM: int = 0x400000cf
SDLK_KP_MEMSTORE: int = 0x400000d0
SDLK_KP_MEMRECALL: int = 0x400000d1
SDLK_KP_MEMCLEAR: int = 0x400000d2
SDLK_KP_MEMADD: int = 0x400000d3
SDLK_KP_MEMSUBTRACT: int = 0x400000d4
SDLK_KP_MEMMULTIPLY: int = 0x400000d5
SDLK_KP_MEMDIVIDE: int = 0x400000d6
SDLK_KP_PLUSMINUS: int = 0x400000d7
SDLK_KP_CLEAR: int = 0x400000d8
SDLK_KP_CLEARENTRY: int = 0x400000d9
SDLK_KP_BINARY: int = 0x400000da
SDLK_KP_OCTAL: int = 0x400000db
SDLK_KP_DECIMAL: int = 0x400000dc
SDLK_KP_HEXADECIMAL: int = 0x400000dd
SDLK_LCTRL: int = 0x400000e0
SDLK_LSHIFT: int = 0x400000e1
SDLK_LALT: int = 0x400000e2
SDLK_LGUI: int = 0x400000e3
SDLK_RCTRL: int = 0x400000e4
SDLK_RSHIFT: int = 0x400000e5
SDLK_RALT: int = 0x400000e6
SDLK_RGUI: int = 0x400000e7
SDLK_MODE: int = 0x40000101
SDLK_SLEEP: int = 0x40000102
SDLK_WAKE: int = 0x40000103
SDLK_CHANNEL_INCREMENT: int = 0x40000104
SDLK_CHANNEL_DECREMENT: int = 0x40000105
SDLK_MEDIA_PLAY: int = 0x40000106
SDLK_MEDIA_PAUSE: int = 0x40000107
SDLK_MEDIA_RECORD: int = 0x40000108
SDLK_MEDIA_FAST_FORWARD: int = 0x40000109
SDLK_MEDIA_REWIND: int = 0x4000010a
SDLK_MEDIA_NEXT_TRACK: int = 0x4000010b
SDLK_MEDIA_PREVIOUS_TRACK: int = 0x4000010c
SDLK_MEDIA_STOP: int = 0x4000010d
SDLK_MEDIA_EJECT: int = 0x4000010e
SDLK_MEDIA_PLAY_PAUSE: int = 0x4000010f
SDLK_MEDIA_SELECT: int = 0x40000110
SDLK_AC_NEW: int = 0x40000111
SDLK_AC_OPEN: int = 0x40000112
SDLK_AC_CLOSE: int = 0x40000113
SDLK_AC_EXIT: int = 0x40000114
SDLK_AC_SAVE: int = 0x40000115
SDLK_AC_PRINT: int = 0x40000116
SDLK_AC_PROPERTIES: int = 0x40000117
SDLK_AC_SEARCH: int = 0x40000118
SDLK_AC_HOME: int = 0x40000119
SDLK_AC_BACK: int = 0x4000011a
SDLK_AC_FORWARD: int = 0x4000011b
SDLK_AC_STOP: int = 0x4000011c
SDLK_AC_REFRESH: int = 0x4000011d
SDLK_AC_BOOKMARKS: int = 0x4000011e
SDLK_SOFTLEFT: int = 0x4000011f
SDLK_SOFTRIGHT: int = 0x40000120
SDLK_CALL: int = 0x40000121
SDLK_ENDCALL: int = 0x40000122
SDLK_LEFT_TAB: int = 0x20000001
SDLK_LEVEL5_SHIFT: int = 0x20000002
SDLK_MULTI_KEY_COMPOSE: int = 0x20000003
SDLK_LMETA: int = 0x20000004
SDLK_RMETA: int = 0x20000005
SDLK_LHYPER: int = 0x20000006
SDLK_RHYPER: int = 0x20000007

SDL_Keymod: typing.TypeAlias = SDL_TYPE["SDL_Keymod", ctypes.c_uint16]

SDL_KMOD_NONE: int = 0x0000
SDL_KMOD_LSHIFT: int = 0x0001
SDL_KMOD_RSHIFT: int = 0x0002
SDL_KMOD_LEVEL5: int = 0x0004
SDL_KMOD_LCTRL: int = 0x0040
SDL_KMOD_RCTRL: int = 0x0080
SDL_KMOD_LALT: int = 0x0100
SDL_KMOD_RALT: int = 0x0200
SDL_KMOD_LGUI: int = 0x0400
SDL_KMOD_RGUI: int = 0x0800
SDL_KMOD_NUM: int = 0x1000
SDL_KMOD_CAPS: int = 0x2000
SDL_KMOD_MODE: int = 0x4000
SDL_KMOD_SCROLL: int = 0x8000
SDL_KMOD_CTRL = SDL_KMOD_LCTRL | SDL_KMOD_RCTRL
SDL_KMOD_SHIFT = SDL_KMOD_LSHIFT | SDL_KMOD_RSHIFT
SDL_KMOD_ALT = SDL_KMOD_LALT | SDL_KMOD_RALT
SDL_KMOD_GUI = SDL_KMOD_LGUI | SDL_KMOD_RGUI


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

class SDL_hid_device(ctypes.c_void_p):
    ...

SDL_hid_bus_type: typing.TypeAlias = SDL_TYPE["SDL_hid_bus_type", SDL_ENUM]

SDL_HID_API_BUS_UNKNOWN, SDL_HID_API_BUS_USB, SDL_HID_API_BUS_BLUETOOTH, \
    SDL_HID_API_BUS_I2C, SDL_HID_API_BUS_SPI = range(5)

SDL_hid_device_info: typing.TypeAlias = SDL_TYPE["SDL_hid_device_info", ctypes.c_void_p]

class SDL_hid_device_info(ctypes.Structure):
    _fields_ = [
        ("path", ctypes.c_char_p),
        ("vendor_id", ctypes.c_ushort),
        ("product_id", ctypes.c_ushort),
        ("serial_number", ctypes.c_wchar_p),
        ("release_number", ctypes.c_ushort),
        ("manufacturer_string", ctypes.c_wchar_p),
        ("product_string", ctypes.c_wchar_p),
        ("usage_page", ctypes.c_ushort),
        ("usage", ctypes.c_ushort),
        ("interface_number", ctypes.c_int),
        ("interface_class", ctypes.c_int),
        ("interface_subclass", ctypes.c_int),
        ("interface_protocol", ctypes.c_int),
        ("bus_type", SDL_hid_bus_type),
        ("next", SDL_POINTER[SDL_hid_device_info])
    ]

SDL_hid_init: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_init", ctypes.c_int, [], SDL_BINARY]
SDL_hid_exit: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_exit", ctypes.c_int, [], SDL_BINARY]
SDL_hid_device_change_count: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_device_change_count", ctypes.c_uint32, [], SDL_BINARY]
SDL_hid_enumerate: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_enumerate", SDL_POINTER[SDL_hid_device_info], [ctypes.c_ushort, ctypes.c_ushort], SDL_BINARY]
SDL_hid_free_enumeration: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_free_enumeration", None, [SDL_POINTER[SDL_hid_device_info]], SDL_BINARY]
SDL_hid_open: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_open", SDL_POINTER[SDL_hid_device], [ctypes.c_ushort, ctypes.c_ushort, ctypes.c_wchar_p], SDL_BINARY]
SDL_hid_open_path: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_open_path", SDL_POINTER[SDL_hid_device], [ctypes.c_char_p], SDL_BINARY]
SDL_hid_write: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_write", ctypes.c_int, [SDL_POINTER[SDL_hid_device], SDL_POINTER[ctypes.c_ubyte], ctypes.c_size_t], SDL_BINARY]
SDL_hid_read_timeout: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_read_timeout", ctypes.c_int, [SDL_POINTER[SDL_hid_device], SDL_POINTER[ctypes.c_ubyte], ctypes.c_size_t, ctypes.c_int], SDL_BINARY]
SDL_hid_read: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_read", ctypes.c_int, [SDL_POINTER[SDL_hid_device], SDL_POINTER[ctypes.c_ubyte], ctypes.c_size_t], SDL_BINARY]
SDL_hid_set_nonblocking: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_set_nonblocking", ctypes.c_int, [SDL_POINTER[SDL_hid_device], ctypes.c_int], SDL_BINARY]
SDL_hid_send_feature_report: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_send_feature_report", ctypes.c_int, [SDL_POINTER[SDL_hid_device], SDL_POINTER[ctypes.c_ubyte], ctypes.c_size_t], SDL_BINARY]
SDL_hid_get_feature_report: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_feature_report", ctypes.c_int, [SDL_POINTER[SDL_hid_device], SDL_POINTER[ctypes.c_ubyte], ctypes.c_size_t], SDL_BINARY]
SDL_hid_get_input_report: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_input_report", ctypes.c_int, [SDL_POINTER[SDL_hid_device], SDL_POINTER[ctypes.c_ubyte], ctypes.c_size_t], SDL_BINARY]
SDL_hid_close: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_close", ctypes.c_int, [SDL_POINTER[SDL_hid_device]], SDL_BINARY]
SDL_hid_get_manufacturer_string: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_manufacturer_string", ctypes.c_int, [SDL_POINTER[SDL_hid_device], ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_hid_get_product_string: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_product_string", ctypes.c_int, [SDL_POINTER[SDL_hid_device], ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_hid_get_serial_number_string: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_serial_number_string", ctypes.c_int, [SDL_POINTER[SDL_hid_device], ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_hid_get_indexed_string: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_indexed_string", ctypes.c_int, [SDL_POINTER[SDL_hid_device], ctypes.c_int, ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_hid_get_device_info: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_device_info", SDL_POINTER[SDL_hid_device_info], [SDL_POINTER[SDL_hid_device]], SDL_BINARY]
SDL_hid_get_report_descriptor: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_get_report_descriptor", ctypes.c_int, [SDL_POINTER[SDL_hid_device], SDL_POINTER[ctypes.c_ubyte], ctypes.c_size_t], SDL_BINARY]
SDL_hid_ble_scan: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_hid_ble_scan", None, [ctypes.c_bool], SDL_BINARY]


from .__init__ import sdl3, ctypes, TEST_RegisterFunction

@TEST_RegisterFunction(["Darwin", "Windows"])
def TEST_SDL_GetPreferredLocales() -> None:
    assert sdl3.SDL_Init(0), sdl3.SDL_GetError().decode()
    count = ctypes.pointer(ctypes.c_int(0))
    locales = sdl3.SDL_GetPreferredLocales(count)
    assert locales, sdl3.SDL_GetError().decode()
    assert count.contents.value != 0, "Failed to get preferred locales."
    sdl3.SDL_Quit()


from .__init__ import ctypes, typing, abc, SDL_POINTER, \
    SDL_FUNC_TYPE, SDL_FUNC, SDL_TYPE, SDL_BINARY, SDL_ENUM

from .SDL_events import SDL_Event

SDL_InitFlags: typing.TypeAlias = SDL_TYPE["SDL_InitFlags", ctypes.c_uint32]

SDL_INIT_AUDIO: int = 0x00000010
SDL_INIT_VIDEO: int = 0x00000020
SDL_INIT_JOYSTICK: int = 0x00000200
SDL_INIT_HAPTIC: int = 0x00001000
SDL_INIT_GAMEPAD: int= 0x00002000
SDL_INIT_EVENTS: int = 0x00004000
SDL_INIT_SENSOR: int = 0x00008000
SDL_INIT_CAMERA: int = 0x00010000

SDL_AppResult: typing.TypeAlias = SDL_TYPE["SDL_AppResult", SDL_ENUM]

SDL_APP_CONTINUE, SDL_APP_SUCCESS, SDL_APP_FAILURE = range(3)

SDL_AppInit_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_AppInit_func", SDL_AppResult, [SDL_POINTER[ctypes.c_void_p], ctypes.c_int, SDL_POINTER[ctypes.c_char_p]]]
SDL_AppIterate_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_AppIterate_func", SDL_AppResult, [ctypes.c_void_p]]
SDL_AppEvent_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_AppEvent_func", SDL_AppResult, [ctypes.c_void_p, SDL_POINTER[SDL_Event]]]
SDL_AppQuit_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_AppQuit_func", None, [ctypes.c_void_p, SDL_AppResult]]

SDL_Init: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_Init", ctypes.c_bool, [SDL_InitFlags], SDL_BINARY]
SDL_InitSubSystem: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_InitSubSystem", ctypes.c_bool, [SDL_InitFlags], SDL_BINARY]
SDL_QuitSubSystem: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_QuitSubSystem", None, [SDL_InitFlags], SDL_BINARY]
SDL_WasInit: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_WasInit", SDL_InitFlags, [SDL_InitFlags], SDL_BINARY]
SDL_Quit: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_Quit", None, [], SDL_BINARY]

SDL_IsMainThread: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_IsMainThread", ctypes.c_bool, [], SDL_BINARY]
SDL_MainThreadCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_MainThreadCallback", None, [ctypes.c_void_p]]
SDL_RunOnMainThread: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RunOnMainThread", ctypes.c_bool, [SDL_MainThreadCallback, ctypes.c_void_p, ctypes.c_bool], SDL_BINARY]

SDL_SetAppMetadata: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAppMetadata", ctypes.c_bool, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_SetAppMetadataProperty: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetAppMetadataProperty", ctypes.c_bool, [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]

SDL_PROP_APP_METADATA_NAME_STRING: bytes = "SDL.app.metadata.name".encode()
SDL_PROP_APP_METADATA_VERSION_STRING: bytes = "SDL.app.metadata.version".encode()
SDL_PROP_APP_METADATA_IDENTIFIER_STRING: bytes = "SDL.app.metadata.identifier".encode()
SDL_PROP_APP_METADATA_CREATOR_STRING: bytes = "SDL.app.metadata.creator".encode()
SDL_PROP_APP_METADATA_COPYRIGHT_STRING: bytes = "SDL.app.metadata.copyright".encode()
SDL_PROP_APP_METADATA_URL_STRING: bytes = "SDL.app.metadata.url".encode()
SDL_PROP_APP_METADATA_TYPE_STRING: bytes = "SDL.app.metadata.type".encode()

SDL_GetAppMetadataProperty: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetAppMetadataProperty", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

SDL_PowerState: typing.TypeAlias = SDL_TYPE["SDL_PowerState", SDL_ENUM]

SDL_POWERSTATE_ERROR, SDL_POWERSTATE_UNKNOWN, SDL_POWERSTATE_ON_BATTERY, \
    SDL_POWERSTATE_NO_BATTERY, SDL_POWERSTATE_CHARGING, SDL_POWERSTATE_CHARGED = range(-1, 5)

SDL_GetPowerInfo: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPowerInfo", SDL_PowerState, [SDL_POINTER[ctypes.c_int], SDL_POINTER[ctypes.c_int]], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_FUNC_TYPE, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

SDL_HINT_ALLOW_ALT_TAB_WHILE_GRABBED: bytes = "SDL_ALLOW_ALT_TAB_WHILE_GRABBED".encode()
SDL_HINT_ANDROID_ALLOW_RECREATE_ACTIVITY: bytes = "SDL_ANDROID_ALLOW_RECREATE_ACTIVITY".encode()
SDL_HINT_ANDROID_BLOCK_ON_PAUSE: bytes = "SDL_ANDROID_BLOCK_ON_PAUSE".encode()
SDL_HINT_ANDROID_LOW_LATENCY_AUDIO: bytes = "SDL_ANDROID_LOW_LATENCY_AUDIO".encode()
SDL_HINT_ANDROID_TRAP_BACK_BUTTON: bytes = "SDL_ANDROID_TRAP_BACK_BUTTON".encode()
SDL_HINT_APP_ID: bytes = "SDL_APP_ID".encode()
SDL_HINT_APP_NAME: bytes = "SDL_APP_NAME".encode()
SDL_HINT_APPLE_TV_CONTROLLER_UI_EVENTS: bytes = "SDL_APPLE_TV_CONTROLLER_UI_EVENTS".encode()
SDL_HINT_APPLE_TV_REMOTE_ALLOW_ROTATION: bytes = "SDL_APPLE_TV_REMOTE_ALLOW_ROTATION".encode()
SDL_HINT_AUDIO_ALSA_DEFAULT_DEVICE: bytes = "SDL_AUDIO_ALSA_DEFAULT_DEVICE".encode()
SDL_HINT_AUDIO_ALSA_DEFAULT_PLAYBACK_DEVICE: bytes = "SDL_AUDIO_ALSA_DEFAULT_PLAYBACK_DEVICE".encode()
SDL_HINT_AUDIO_ALSA_DEFAULT_RECORDING_DEVICE: bytes = "SDL_AUDIO_ALSA_DEFAULT_RECORDING_DEVICE".encode()
SDL_HINT_AUDIO_CATEGORY: bytes = "SDL_AUDIO_CATEGORY".encode()
SDL_HINT_AUDIO_CHANNELS: bytes = "SDL_AUDIO_CHANNELS".encode()
SDL_HINT_AUDIO_DEVICE_APP_ICON_NAME: bytes = "SDL_AUDIO_DEVICE_APP_ICON_NAME".encode()
SDL_HINT_AUDIO_DEVICE_SAMPLE_FRAMES: bytes = "SDL_AUDIO_DEVICE_SAMPLE_FRAMES".encode()
SDL_HINT_AUDIO_DEVICE_STREAM_NAME: bytes = "SDL_AUDIO_DEVICE_STREAM_NAME".encode()
SDL_HINT_AUDIO_DEVICE_STREAM_ROLE: bytes = "SDL_AUDIO_DEVICE_STREAM_ROLE".encode()
SDL_HINT_AUDIO_DISK_INPUT_FILE: bytes = "SDL_AUDIO_DISK_INPUT_FILE".encode()
SDL_HINT_AUDIO_DISK_OUTPUT_FILE: bytes = "SDL_AUDIO_DISK_OUTPUT_FILE".encode()
SDL_HINT_AUDIO_DISK_TIMESCALE: bytes = "SDL_AUDIO_DISK_TIMESCALE".encode()
SDL_HINT_AUDIO_DRIVER: bytes = "SDL_AUDIO_DRIVER".encode()
SDL_HINT_AUDIO_DUMMY_TIMESCALE: bytes = "SDL_AUDIO_DUMMY_TIMESCALE".encode()
SDL_HINT_AUDIO_FORMAT: bytes = "SDL_AUDIO_FORMAT".encode()
SDL_HINT_AUDIO_FREQUENCY: bytes = "SDL_AUDIO_FREQUENCY".encode()
SDL_HINT_AUDIO_INCLUDE_MONITORS: bytes = "SDL_AUDIO_INCLUDE_MONITORS".encode()
SDL_HINT_AUTO_UPDATE_JOYSTICKS: bytes = "SDL_AUTO_UPDATE_JOYSTICKS".encode()
SDL_HINT_AUTO_UPDATE_SENSORS: bytes = "SDL_AUTO_UPDATE_SENSORS".encode()
SDL_HINT_BMP_SAVE_LEGACY_FORMAT: bytes = "SDL_BMP_SAVE_LEGACY_FORMAT".encode()
SDL_HINT_CAMERA_DRIVER: bytes = "SDL_CAMERA_DRIVER".encode()
SDL_HINT_CPU_FEATURE_MASK: bytes = "SDL_CPU_FEATURE_MASK".encode()
SDL_HINT_JOYSTICK_DIRECTINPUT: bytes = "SDL_JOYSTICK_DIRECTINPUT".encode()
SDL_HINT_FILE_DIALOG_DRIVER: bytes = "SDL_FILE_DIALOG_DRIVER".encode()
SDL_HINT_DISPLAY_USABLE_BOUNDS: bytes = "SDL_DISPLAY_USABLE_BOUNDS".encode()
SDL_HINT_EMSCRIPTEN_ASYNCIFY: bytes = "SDL_EMSCRIPTEN_ASYNCIFY".encode()
SDL_HINT_EMSCRIPTEN_CANVAS_SELECTOR: bytes = "SDL_EMSCRIPTEN_CANVAS_SELECTOR".encode()
SDL_HINT_EMSCRIPTEN_KEYBOARD_ELEMENT: bytes = "SDL_EMSCRIPTEN_KEYBOARD_ELEMENT".encode()
SDL_HINT_ENABLE_SCREEN_KEYBOARD: bytes = "SDL_ENABLE_SCREEN_KEYBOARD".encode()
SDL_HINT_EVDEV_DEVICES: bytes = "SDL_EVDEV_DEVICES".encode()
SDL_HINT_EVENT_LOGGING: bytes = "SDL_EVENT_LOGGING".encode()
SDL_HINT_FORCE_RAISEWINDOW: bytes = "SDL_FORCE_RAISEWINDOW".encode()
SDL_HINT_FRAMEBUFFER_ACCELERATION: bytes = "SDL_FRAMEBUFFER_ACCELERATION".encode()
SDL_HINT_GAMECONTROLLERCONFIG: bytes = "SDL_GAMECONTROLLERCONFIG".encode()
SDL_HINT_GAMECONTROLLERCONFIG_FILE: bytes = "SDL_GAMECONTROLLERCONFIG_FILE".encode()
SDL_HINT_GAMECONTROLLERTYPE: bytes = "SDL_GAMECONTROLLERTYPE".encode()
SDL_HINT_GAMECONTROLLER_IGNORE_DEVICES: bytes = "SDL_GAMECONTROLLER_IGNORE_DEVICES".encode()
SDL_HINT_GAMECONTROLLER_IGNORE_DEVICES_EXCEPT: bytes = "SDL_GAMECONTROLLER_IGNORE_DEVICES_EXCEPT".encode()
SDL_HINT_GAMECONTROLLER_SENSOR_FUSION: bytes = "SDL_GAMECONTROLLER_SENSOR_FUSION".encode()
SDL_HINT_GDK_TEXTINPUT_DEFAULT_TEXT: bytes = "SDL_GDK_TEXTINPUT_DEFAULT_TEXT".encode()
SDL_HINT_GDK_TEXTINPUT_DESCRIPTION: bytes = "SDL_GDK_TEXTINPUT_DESCRIPTION".encode()
SDL_HINT_GDK_TEXTINPUT_MAX_LENGTH: bytes = "SDL_GDK_TEXTINPUT_MAX_LENGTH".encode()
SDL_HINT_GDK_TEXTINPUT_SCOPE: bytes = "SDL_GDK_TEXTINPUT_SCOPE".encode()
SDL_HINT_GDK_TEXTINPUT_TITLE: bytes = "SDL_GDK_TEXTINPUT_TITLE".encode()
SDL_HINT_HIDAPI_LIBUSB: bytes = "SDL_HIDAPI_LIBUSB".encode()
SDL_HINT_HIDAPI_LIBUSB_WHITELIST: bytes = "SDL_HIDAPI_LIBUSB_WHITELIST".encode()
SDL_HINT_HIDAPI_UDEV: bytes = "SDL_HIDAPI_UDEV".encode()
SDL_HINT_GPU_DRIVER: bytes = "SDL_GPU_DRIVER".encode()
SDL_HINT_HIDAPI_ENUMERATE_ONLY_CONTROLLERS: bytes = "SDL_HIDAPI_ENUMERATE_ONLY_CONTROLLERS".encode()
SDL_HINT_HIDAPI_IGNORE_DEVICES: bytes = "SDL_HIDAPI_IGNORE_DEVICES".encode()
SDL_HINT_IME_IMPLEMENTED_UI: bytes = "SDL_IME_IMPLEMENTED_UI".encode()
SDL_HINT_IOS_HIDE_HOME_INDICATOR: bytes = "SDL_IOS_HIDE_HOME_INDICATOR".encode()
SDL_HINT_JOYSTICK_ALLOW_BACKGROUND_EVENTS: bytes = "SDL_JOYSTICK_ALLOW_BACKGROUND_EVENTS".encode()
SDL_HINT_JOYSTICK_ARCADESTICK_DEVICES: bytes = "SDL_JOYSTICK_ARCADESTICK_DEVICES".encode()
SDL_HINT_JOYSTICK_ARCADESTICK_DEVICES_EXCLUDED: bytes = "SDL_JOYSTICK_ARCADESTICK_DEVICES_EXCLUDED".encode()
SDL_HINT_JOYSTICK_BLACKLIST_DEVICES: bytes = "SDL_JOYSTICK_BLACKLIST_DEVICES".encode()
SDL_HINT_JOYSTICK_BLACKLIST_DEVICES_EXCLUDED: bytes = "SDL_JOYSTICK_BLACKLIST_DEVICES_EXCLUDED".encode()
SDL_HINT_JOYSTICK_DEVICE: bytes = "SDL_JOYSTICK_DEVICE".encode()
SDL_HINT_JOYSTICK_ENHANCED_REPORTS: bytes = "SDL_JOYSTICK_ENHANCED_REPORTS".encode()
SDL_HINT_JOYSTICK_FLIGHTSTICK_DEVICES: bytes = "SDL_JOYSTICK_FLIGHTSTICK_DEVICES".encode()
SDL_HINT_JOYSTICK_FLIGHTSTICK_DEVICES_EXCLUDED: bytes = "SDL_JOYSTICK_FLIGHTSTICK_DEVICES_EXCLUDED".encode()
SDL_HINT_JOYSTICK_GAMEINPUT: bytes = "SDL_JOYSTICK_GAMEINPUT".encode()
SDL_HINT_JOYSTICK_GAMECUBE_DEVICES: bytes = "SDL_JOYSTICK_GAMECUBE_DEVICES".encode()
SDL_HINT_JOYSTICK_GAMECUBE_DEVICES_EXCLUDED: bytes = "SDL_JOYSTICK_GAMECUBE_DEVICES_EXCLUDED".encode()
SDL_HINT_JOYSTICK_HIDAPI: bytes = "SDL_JOYSTICK_HIDAPI".encode()
SDL_HINT_JOYSTICK_HIDAPI_COMBINE_JOY_CONS: bytes = "SDL_JOYSTICK_HIDAPI_COMBINE_JOY_CONS".encode()
SDL_HINT_JOYSTICK_HIDAPI_GAMECUBE: bytes = "SDL_JOYSTICK_HIDAPI_GAMECUBE".encode()
SDL_HINT_JOYSTICK_HIDAPI_GAMECUBE_RUMBLE_BRAKE: bytes = "SDL_JOYSTICK_HIDAPI_GAMECUBE_RUMBLE_BRAKE".encode()
SDL_HINT_JOYSTICK_HIDAPI_JOY_CONS: bytes = "SDL_JOYSTICK_HIDAPI_JOY_CONS".encode()
SDL_HINT_JOYSTICK_HIDAPI_JOYCON_HOME_LED: bytes = "SDL_JOYSTICK_HIDAPI_JOYCON_HOME_LED".encode()
SDL_HINT_JOYSTICK_HIDAPI_LUNA: bytes = "SDL_JOYSTICK_HIDAPI_LUNA".encode()
SDL_HINT_JOYSTICK_HIDAPI_NINTENDO_CLASSIC: bytes = "SDL_JOYSTICK_HIDAPI_NINTENDO_CLASSIC".encode()
SDL_HINT_JOYSTICK_HIDAPI_PS3: bytes = "SDL_JOYSTICK_HIDAPI_PS3".encode()
SDL_HINT_JOYSTICK_HIDAPI_PS3_SIXAXIS_DRIVER: bytes = "SDL_JOYSTICK_HIDAPI_PS3_SIXAXIS_DRIVER".encode()
SDL_HINT_JOYSTICK_HIDAPI_PS4: bytes = "SDL_JOYSTICK_HIDAPI_PS4".encode()
SDL_HINT_JOYSTICK_HIDAPI_PS4_REPORT_INTERVAL: bytes = "SDL_JOYSTICK_HIDAPI_PS4_REPORT_INTERVAL".encode()
SDL_HINT_JOYSTICK_HIDAPI_PS5: bytes = "SDL_JOYSTICK_HIDAPI_PS5".encode()
SDL_HINT_JOYSTICK_HIDAPI_PS5_PLAYER_LED: bytes = "SDL_JOYSTICK_HIDAPI_PS5_PLAYER_LED".encode()
SDL_HINT_JOYSTICK_HIDAPI_SHIELD: bytes = "SDL_JOYSTICK_HIDAPI_SHIELD".encode()
SDL_HINT_JOYSTICK_HIDAPI_STADIA: bytes = "SDL_JOYSTICK_HIDAPI_STADIA".encode()
SDL_HINT_JOYSTICK_HIDAPI_STEAM: bytes = "SDL_JOYSTICK_HIDAPI_STEAM".encode()
SDL_HINT_JOYSTICK_HIDAPI_STEAM_HOME_LED: bytes = "SDL_JOYSTICK_HIDAPI_STEAM_HOME_LED".encode()
SDL_HINT_JOYSTICK_HIDAPI_STEAMDECK: bytes = "SDL_JOYSTICK_HIDAPI_STEAMDECK".encode()
SDL_HINT_JOYSTICK_HIDAPI_STEAM_HORI: bytes = "SDL_JOYSTICK_HIDAPI_STEAM_HORI".encode()
SDL_HINT_JOYSTICK_HIDAPI_SWITCH: bytes = "SDL_JOYSTICK_HIDAPI_SWITCH".encode()
SDL_HINT_JOYSTICK_HIDAPI_SWITCH_HOME_LED: bytes = "SDL_JOYSTICK_HIDAPI_SWITCH_HOME_LED".encode()
SDL_HINT_JOYSTICK_HIDAPI_SWITCH_PLAYER_LED: bytes = "SDL_JOYSTICK_HIDAPI_SWITCH_PLAYER_LED".encode()
SDL_HINT_JOYSTICK_HIDAPI_VERTICAL_JOY_CONS: bytes = "SDL_JOYSTICK_HIDAPI_VERTICAL_JOY_CONS".encode()
SDL_HINT_JOYSTICK_HIDAPI_WII: bytes = "SDL_JOYSTICK_HIDAPI_WII".encode()
SDL_HINT_JOYSTICK_HIDAPI_WII_PLAYER_LED: bytes = "SDL_JOYSTICK_HIDAPI_WII_PLAYER_LED".encode()
SDL_HINT_JOYSTICK_HIDAPI_XBOX: bytes = "SDL_JOYSTICK_HIDAPI_XBOX".encode()
SDL_HINT_JOYSTICK_HIDAPI_XBOX_360: bytes = "SDL_JOYSTICK_HIDAPI_XBOX_360".encode()
SDL_HINT_JOYSTICK_HIDAPI_XBOX_360_PLAYER_LED: bytes = "SDL_JOYSTICK_HIDAPI_XBOX_360_PLAYER_LED".encode()
SDL_HINT_JOYSTICK_HIDAPI_XBOX_360_WIRELESS: bytes = "SDL_JOYSTICK_HIDAPI_XBOX_360_WIRELESS".encode()
SDL_HINT_JOYSTICK_HIDAPI_XBOX_ONE: bytes = "SDL_JOYSTICK_HIDAPI_XBOX_ONE".encode()
SDL_HINT_JOYSTICK_HIDAPI_XBOX_ONE_HOME_LED: bytes = "SDL_JOYSTICK_HIDAPI_XBOX_ONE_HOME_LED".encode()
SDL_HINT_JOYSTICK_IOKIT: bytes = "SDL_JOYSTICK_IOKIT".encode()
SDL_HINT_JOYSTICK_LINUX_CLASSIC: bytes = "SDL_JOYSTICK_LINUX_CLASSIC".encode()
SDL_HINT_JOYSTICK_LINUX_DEADZONES: bytes = "SDL_JOYSTICK_LINUX_DEADZONES".encode()
SDL_HINT_JOYSTICK_LINUX_DIGITAL_HATS: bytes = "SDL_JOYSTICK_LINUX_DIGITAL_HATS".encode()
SDL_HINT_JOYSTICK_LINUX_HAT_DEADZONES: bytes = "SDL_JOYSTICK_LINUX_HAT_DEADZONES".encode()
SDL_HINT_JOYSTICK_MFI: bytes = "SDL_JOYSTICK_MFI".encode()
SDL_HINT_JOYSTICK_RAWINPUT: bytes = "SDL_JOYSTICK_RAWINPUT".encode()
SDL_HINT_JOYSTICK_RAWINPUT_CORRELATE_XINPUT: bytes = "SDL_JOYSTICK_RAWINPUT_CORRELATE_XINPUT".encode()
SDL_HINT_JOYSTICK_ROG_CHAKRAM: bytes = "SDL_JOYSTICK_ROG_CHAKRAM".encode()
SDL_HINT_JOYSTICK_THREAD: bytes = "SDL_JOYSTICK_THREAD".encode()
SDL_HINT_JOYSTICK_THROTTLE_DEVICES: bytes = "SDL_JOYSTICK_THROTTLE_DEVICES".encode()
SDL_HINT_JOYSTICK_THROTTLE_DEVICES_EXCLUDED: bytes = "SDL_JOYSTICK_THROTTLE_DEVICES_EXCLUDED".encode()
SDL_HINT_JOYSTICK_WGI: bytes = "SDL_JOYSTICK_WGI".encode()
SDL_HINT_JOYSTICK_WHEEL_DEVICES: bytes = "SDL_JOYSTICK_WHEEL_DEVICES".encode()
SDL_HINT_JOYSTICK_WHEEL_DEVICES_EXCLUDED: bytes = "SDL_JOYSTICK_WHEEL_DEVICES_EXCLUDED".encode()
SDL_HINT_JOYSTICK_ZERO_CENTERED_DEVICES: bytes = "SDL_JOYSTICK_ZERO_CENTERED_DEVICES".encode()
SDL_HINT_JOYSTICK_HAPTIC_AXES: bytes = "SDL_JOYSTICK_HAPTIC_AXES".encode()
SDL_HINT_KEYCODE_OPTIONS: bytes = "SDL_KEYCODE_OPTIONS".encode()
SDL_HINT_KMSDRM_DEVICE_INDEX: bytes = "SDL_KMSDRM_DEVICE_INDEX".encode()
SDL_HINT_KMSDRM_REQUIRE_DRM_MASTER: bytes = "SDL_KMSDRM_REQUIRE_DRM_MASTER".encode()
SDL_HINT_LOGGING: bytes = "SDL_LOGGING".encode()
SDL_HINT_MAC_BACKGROUND_APP: bytes = "SDL_MAC_BACKGROUND_APP".encode()
SDL_HINT_MAC_CTRL_CLICK_EMULATE_RIGHT_CLICK: bytes = "SDL_MAC_CTRL_CLICK_EMULATE_RIGHT_CLICK".encode()
SDL_HINT_MAC_OPENGL_ASYNC_DISPATCH: bytes = "SDL_MAC_OPENGL_ASYNC_DISPATCH".encode()
SDL_HINT_MAC_OPTION_AS_ALT: bytes = "SDL_MAC_OPTION_AS_ALT".encode()
SDL_HINT_MAC_SCROLL_MOMENTUM: bytes = "SDL_MAC_SCROLL_MOMENTUM".encode()
SDL_HINT_MAIN_CALLBACK_RATE: bytes = "SDL_MAIN_CALLBACK_RATE".encode()
SDL_HINT_MOUSE_AUTO_CAPTURE: bytes = "SDL_MOUSE_AUTO_CAPTURE".encode()
SDL_HINT_MOUSE_DOUBLE_CLICK_RADIUS: bytes = "SDL_MOUSE_DOUBLE_CLICK_RADIUS".encode()
SDL_HINT_MOUSE_DOUBLE_CLICK_TIME: bytes = "SDL_MOUSE_DOUBLE_CLICK_TIME".encode()
SDL_HINT_MOUSE_DEFAULT_SYSTEM_CURSOR: bytes = "SDL_MOUSE_DEFAULT_SYSTEM_CURSOR".encode()
SDL_HINT_MOUSE_EMULATE_WARP_WITH_RELATIVE: bytes = "SDL_MOUSE_EMULATE_WARP_WITH_RELATIVE".encode()
SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH: bytes = "SDL_MOUSE_FOCUS_CLICKTHROUGH".encode()
SDL_HINT_MOUSE_NORMAL_SPEED_SCALE: bytes = "SDL_MOUSE_NORMAL_SPEED_SCALE".encode()
SDL_HINT_MOUSE_RELATIVE_MODE_CENTER: bytes = "SDL_MOUSE_RELATIVE_MODE_CENTER".encode()
SDL_HINT_MOUSE_RELATIVE_SPEED_SCALE: bytes = "SDL_MOUSE_RELATIVE_SPEED_SCALE".encode()
SDL_HINT_MOUSE_RELATIVE_SYSTEM_SCALE: bytes = "SDL_MOUSE_RELATIVE_SYSTEM_SCALE".encode()
SDL_HINT_MOUSE_RELATIVE_WARP_MOTION: bytes = "SDL_MOUSE_RELATIVE_WARP_MOTION".encode()
SDL_HINT_MOUSE_RELATIVE_CURSOR_VISIBLE: bytes = "SDL_MOUSE_RELATIVE_CURSOR_VISIBLE".encode()
SDL_HINT_MOUSE_TOUCH_EVENTS: bytes = "SDL_MOUSE_TOUCH_EVENTS".encode()
SDL_HINT_MUTE_CONSOLE_KEYBOARD: bytes = "SDL_MUTE_CONSOLE_KEYBOARD".encode()
SDL_HINT_NO_SIGNAL_HANDLERS: bytes = "SDL_NO_SIGNAL_HANDLERS".encode()
SDL_HINT_OPENGL_LIBRARY: bytes = "SDL_OPENGL_LIBRARY".encode()
SDL_HINT_EGL_LIBRARY: bytes = "SDL_OPENGL_LIBRARY".encode()
SDL_HINT_OPENGL_ES_DRIVER: bytes = "SDL_OPENGL_ES_DRIVER".encode()
SDL_HINT_OPENVR_LIBRARY: bytes = "SDL_OPENVR_LIBRARY".encode()
SDL_HINT_ORIENTATIONS: bytes = "SDL_ORIENTATIONS".encode()
SDL_HINT_PEN_DELAY_MOUSE_BUTTON: bytes = "SDL_PEN_DELAY_MOUSE_BUTTON".encode()
SDL_HINT_PEN_NOT_MOUSE: bytes = "SDL_PEN_NOT_MOUSE".encode()
SDL_HINT_POLL_SENTINEL: bytes = "SDL_POLL_SENTINEL".encode()
SDL_HINT_PREFERRED_LOCALES: bytes = "SDL_PREFERRED_LOCALES".encode()
SDL_HINT_QUIT_ON_LAST_WINDOW_CLOSE: bytes = "SDL_QUIT_ON_LAST_WINDOW_CLOSE".encode()
SDL_HINT_RENDER_DIRECT3D_THREADSAFE: bytes = "SDL_RENDER_DIRECT3D_THREADSAFE".encode()
SDL_HINT_RENDER_DIRECT3D11_DEBUG: bytes = "SDL_RENDER_DIRECT3D11_DEBUG".encode()
SDL_HINT_RENDER_VULKAN_DEBUG: bytes = "SDL_RENDER_VULKAN_DEBUG".encode()
SDL_HINT_RENDER_GPU_DEBUG: bytes = "SDL_RENDER_GPU_DEBUG".encode()
SDL_HINT_RENDER_GPU_LOW_POWER: bytes = "SDL_RENDER_GPU_LOW_POWER".encode()
SDL_HINT_RENDER_DRIVER: bytes = "SDL_RENDER_DRIVER".encode()
SDL_HINT_RENDER_LINE_METHOD: bytes = "SDL_RENDER_LINE_METHOD".encode()
SDL_HINT_RENDER_METAL_PREFER_LOW_POWER_DEVICE: bytes = "SDL_RENDER_METAL_PREFER_LOW_POWER_DEVICE".encode()
SDL_HINT_RENDER_VSYNC: bytes = "SDL_RENDER_VSYNC".encode()
SDL_HINT_RETURN_KEY_HIDES_IME: bytes = "SDL_RETURN_KEY_HIDES_IME".encode()
SDL_HINT_ROG_GAMEPAD_MICE: bytes = "SDL_ROG_GAMEPAD_MICE".encode()
SDL_HINT_ROG_GAMEPAD_MICE_EXCLUDED: bytes = "SDL_ROG_GAMEPAD_MICE_EXCLUDED".encode()
SDL_HINT_RPI_VIDEO_LAYER: bytes = "SDL_RPI_VIDEO_LAYER".encode()
SDL_HINT_SCREENSAVER_INHIBIT_ACTIVITY_NAME: bytes = "SDL_SCREENSAVER_INHIBIT_ACTIVITY_NAME".encode()
SDL_HINT_SHUTDOWN_DBUS_ON_QUIT: bytes = "SDL_SHUTDOWN_DBUS_ON_QUIT".encode()
SDL_HINT_STORAGE_TITLE_DRIVER: bytes = "SDL_STORAGE_TITLE_DRIVER".encode()
SDL_HINT_STORAGE_USER_DRIVER: bytes = "SDL_STORAGE_USER_DRIVER".encode()
SDL_HINT_THREAD_FORCE_REALTIME_TIME_CRITICAL: bytes = "SDL_THREAD_FORCE_REALTIME_TIME_CRITICAL".encode()
SDL_HINT_THREAD_PRIORITY_POLICY: bytes = "SDL_THREAD_PRIORITY_POLICY".encode()
SDL_HINT_TIMER_RESOLUTION: bytes = "SDL_TIMER_RESOLUTION".encode()
SDL_HINT_TOUCH_MOUSE_EVENTS: bytes = "SDL_TOUCH_MOUSE_EVENTS".encode()
SDL_HINT_TRACKPAD_IS_TOUCH_ONLY: bytes = "SDL_TRACKPAD_IS_TOUCH_ONLY".encode()
SDL_HINT_TV_REMOTE_AS_JOYSTICK: bytes = "SDL_TV_REMOTE_AS_JOYSTICK".encode()
SDL_HINT_VIDEO_ALLOW_SCREENSAVER: bytes = "SDL_VIDEO_ALLOW_SCREENSAVER".encode()
SDL_HINT_VIDEO_DISPLAY_PRIORITY: bytes = "SDL_VIDEO_DISPLAY_PRIORITY".encode()
SDL_HINT_VIDEO_DOUBLE_BUFFER: bytes = "SDL_VIDEO_DOUBLE_BUFFER".encode()
SDL_HINT_VIDEO_DRIVER: bytes = "SDL_VIDEO_DRIVER".encode()
SDL_HINT_VIDEO_DUMMY_SAVE_FRAMES: bytes = "SDL_VIDEO_DUMMY_SAVE_FRAMES".encode()
SDL_HINT_VIDEO_EGL_ALLOW_GETDISPLAY_FALLBACK: bytes = "SDL_VIDEO_EGL_ALLOW_GETDISPLAY_FALLBACK".encode()
SDL_HINT_VIDEO_FORCE_EGL: bytes = "SDL_VIDEO_FORCE_EGL".encode()
SDL_HINT_VIDEO_MAC_FULLSCREEN_SPACES: bytes = "SDL_VIDEO_MAC_FULLSCREEN_SPACES".encode()
SDL_HINT_VIDEO_MAC_FULLSCREEN_MENU_VISIBILITY: bytes = "SDL_VIDEO_MAC_FULLSCREEN_MENU_VISIBILITY".encode()
SDL_HINT_VIDEO_MINIMIZE_ON_FOCUS_LOSS: bytes = "SDL_VIDEO_MINIMIZE_ON_FOCUS_LOSS".encode()
SDL_HINT_VIDEO_OFFSCREEN_SAVE_FRAMES: bytes = "SDL_VIDEO_OFFSCREEN_SAVE_FRAMES".encode()
SDL_HINT_VIDEO_SYNC_WINDOW_OPERATIONS: bytes = "SDL_VIDEO_SYNC_WINDOW_OPERATIONS".encode()
SDL_HINT_VIDEO_WAYLAND_ALLOW_LIBDECOR: bytes = "SDL_VIDEO_WAYLAND_ALLOW_LIBDECOR".encode()
SDL_HINT_VIDEO_WAYLAND_MODE_EMULATION: bytes = "SDL_VIDEO_WAYLAND_MODE_EMULATION".encode()
SDL_HINT_VIDEO_WAYLAND_MODE_SCALING: bytes = "SDL_VIDEO_WAYLAND_MODE_SCALING".encode()
SDL_HINT_VIDEO_WAYLAND_PREFER_LIBDECOR: bytes = "SDL_VIDEO_WAYLAND_PREFER_LIBDECOR".encode()
SDL_HINT_VIDEO_WAYLAND_SCALE_TO_DISPLAY: bytes = "SDL_VIDEO_WAYLAND_SCALE_TO_DISPLAY".encode()
SDL_HINT_VIDEO_WIN_D3DCOMPILER: bytes = "SDL_VIDEO_WIN_D3DCOMPILER".encode()
SDL_HINT_VIDEO_X11_EXTERNAL_WINDOW_INPUT: bytes = "SDL_VIDEO_X11_EXTERNAL_WINDOW_INPUT".encode()
SDL_HINT_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR: bytes = "SDL_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR".encode()
SDL_HINT_VIDEO_X11_NET_WM_PING: bytes = "SDL_VIDEO_X11_NET_WM_PING".encode()
SDL_HINT_VIDEO_X11_NODIRECTCOLOR: bytes = "SDL_VIDEO_X11_NODIRECTCOLOR".encode()
SDL_HINT_VIDEO_X11_SCALING_FACTOR: bytes = "SDL_VIDEO_X11_SCALING_FACTOR".encode()
SDL_HINT_VIDEO_X11_VISUALID: bytes = "SDL_VIDEO_X11_VISUALID".encode()
SDL_HINT_VIDEO_X11_WINDOW_VISUALID: bytes = "SDL_VIDEO_X11_WINDOW_VISUALID".encode()
SDL_HINT_VIDEO_X11_XRANDR: bytes = "SDL_VIDEO_X11_XRANDR".encode()
SDL_HINT_VITA_ENABLE_BACK_TOUCH: bytes = "SDL_VITA_ENABLE_BACK_TOUCH".encode()
SDL_HINT_VITA_ENABLE_FRONT_TOUCH: bytes = "SDL_VITA_ENABLE_FRONT_TOUCH".encode()
SDL_HINT_VITA_MODULE_PATH: bytes = "SDL_VITA_MODULE_PATH".encode()
SDL_HINT_VITA_PVR_INIT: bytes = "SDL_VITA_PVR_INIT".encode()
SDL_HINT_VITA_RESOLUTION: bytes = "SDL_VITA_RESOLUTION".encode()
SDL_HINT_VITA_PVR_OPENGL: bytes = "SDL_VITA_PVR_OPENGL".encode()
SDL_HINT_VITA_TOUCH_MOUSE_DEVICE: bytes = "SDL_VITA_TOUCH_MOUSE_DEVICE".encode()
SDL_HINT_VULKAN_DISPLAY: bytes = "SDL_VULKAN_DISPLAY".encode()
SDL_HINT_VULKAN_LIBRARY: bytes = "SDL_VULKAN_LIBRARY".encode()
SDL_HINT_WAVE_FACT_CHUNK: bytes = "SDL_WAVE_FACT_CHUNK".encode()
SDL_HINT_WAVE_CHUNK_LIMIT: bytes = "SDL_WAVE_CHUNK_LIMIT".encode()
SDL_HINT_WAVE_RIFF_CHUNK_SIZE: bytes = "SDL_WAVE_RIFF_CHUNK_SIZE".encode()
SDL_HINT_WAVE_TRUNCATION: bytes = "SDL_WAVE_TRUNCATION".encode()
SDL_HINT_WINDOW_ACTIVATE_WHEN_RAISED: bytes = "SDL_WINDOW_ACTIVATE_WHEN_RAISED".encode()
SDL_HINT_WINDOW_ACTIVATE_WHEN_SHOWN: bytes = "SDL_WINDOW_ACTIVATE_WHEN_SHOWN".encode()
SDL_HINT_WINDOW_ALLOW_TOPMOST: bytes = "SDL_WINDOW_ALLOW_TOPMOST".encode()
SDL_HINT_WINDOW_FRAME_USABLE_WHILE_CURSOR_HIDDEN: bytes = "SDL_WINDOW_FRAME_USABLE_WHILE_CURSOR_HIDDEN".encode()
SDL_HINT_WINDOWS_CLOSE_ON_ALT_F4: bytes = "SDL_WINDOWS_CLOSE_ON_ALT_F4".encode()
SDL_HINT_WINDOWS_ENABLE_MENU_MNEMONICS: bytes = "SDL_WINDOWS_ENABLE_MENU_MNEMONICS".encode()
SDL_HINT_WINDOWS_ENABLE_MESSAGELOOP: bytes = "SDL_WINDOWS_ENABLE_MESSAGELOOP".encode()
SDL_HINT_WINDOWS_GAMEINPUT: bytes = "SDL_WINDOWS_GAMEINPUT".encode()
SDL_HINT_WINDOWS_RAW_KEYBOARD: bytes = "SDL_WINDOWS_RAW_KEYBOARD".encode()
SDL_HINT_WINDOWS_FORCE_SEMAPHORE_KERNEL: bytes = "SDL_WINDOWS_FORCE_SEMAPHORE_KERNEL".encode()
SDL_HINT_WINDOWS_INTRESOURCE_ICON: bytes = "SDL_WINDOWS_INTRESOURCE_ICON".encode()
SDL_HINT_WINDOWS_INTRESOURCE_ICON_SMALL: bytes = "SDL_WINDOWS_INTRESOURCE_ICON_SMALL".encode()
SDL_HINT_WINDOWS_USE_D3D9EX: bytes = "SDL_WINDOWS_USE_D3D9EX".encode()
SDL_HINT_WINDOWS_ERASE_BACKGROUND_MODE: bytes = "SDL_WINDOWS_ERASE_BACKGROUND_MODE".encode()
SDL_HINT_X11_FORCE_OVERRIDE_REDIRECT: bytes = "SDL_X11_FORCE_OVERRIDE_REDIRECT".encode()
SDL_HINT_X11_WINDOW_TYPE: bytes = "SDL_X11_WINDOW_TYPE".encode()
SDL_HINT_X11_XCB_LIBRARY: bytes = "SDL_X11_XCB_LIBRARY".encode()
SDL_HINT_XINPUT_ENABLED: bytes = "SDL_XINPUT_ENABLED".encode()
SDL_HINT_ASSERT: bytes = "SDL_ASSERT".encode()
SDL_HINT_PEN_MOUSE_EVENTS: bytes = "SDL_PEN_MOUSE_EVENTS".encode()
SDL_HINT_PEN_TOUCH_EVENTS: bytes = "SDL_PEN_TOUCH_EVENTS".encode()

SDL_HintPriority: typing.TypeAlias = SDL_TYPE["SDL_HintPriority", SDL_ENUM]

SDL_HINT_DEFAULT, SDL_HINT_NORMAL, SDL_HINT_OVERRIDE = range(3)

SDL_SetHintWithPriority: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetHintWithPriority", ctypes.c_bool, [ctypes.c_char_p, ctypes.c_char_p, SDL_HintPriority], SDL_BINARY]
SDL_SetHint: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetHint", ctypes.c_bool, [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_ResetHint: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ResetHint", ctypes.c_bool, [ctypes.c_char_p], SDL_BINARY]
SDL_ResetHints: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ResetHints", None, [], SDL_BINARY]
SDL_GetHint: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetHint", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]
SDL_GetHintBoolean: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetHintBoolean", ctypes.c_bool, [ctypes.c_char_p, ctypes.c_bool], SDL_BINARY]

SDL_HintCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_HintCallback", None, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]]

SDL_AddHintCallback: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AddHintCallback", ctypes.c_bool, [ctypes.c_char_p, SDL_HintCallback, ctypes.c_void_p], SDL_BINARY]
SDL_RemoveHintCallback: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RemoveHintCallback", None, [ctypes.c_char_p, SDL_HintCallback, ctypes.c_void_p], SDL_BINARY]


"""A pure Python wrapper for SDL3."""

__version__ = "0.9.8b3"

import sys, os, requests, ctypes, ctypes.util, platform, keyword, inspect, collections.abc as abc, \
    asyncio, aiohttp, zipfile, typing, types, array, importlib, atexit, packaging.version, json, re

SDL_BINARY, SDL_IMAGE_BINARY, SDL_MIXER_BINARY, SDL_TTF_BINARY, SDL_RTF_BINARY, SDL_NET_BINARY, SDL_SHADERCROSS_BINARY = \
    "SDL3", "SDL3_image", "SDL3_mixer", "SDL3_ttf", "SDL3_rtf", "SDL3_net", "SDL3_shadercross"

SDL_BINARY_VAR_MAP: dict[str, str] = {}

for i in locals().copy():
    if i.startswith("SDL_") and i.endswith("_BINARY"):
        SDL_BINARY_VAR_MAP[i] = locals()[i]

SDL_BINARY_VAR_MAP_INV: dict[str, str] = {value: key for key, value in SDL_BINARY_VAR_MAP.items()}
SDL_REPOSITORIES: list[str] = [key.replace("3", "") for key, _ in SDL_BINARY_VAR_MAP_INV.items()]

def SDL_FORMAT_ARCH(arch: str) -> str:
    """Format the architecture string."""
    if arch.lower() in ["x64", "x86_64", "amd64"]: return "AMD64"
    if arch.lower() in ["aarch64", "arm64"]: return "ARM64"
    assert False, "Unknown architecture."

SDL_SYSTEM, SDL_ARCH = platform.system(), SDL_FORMAT_ARCH(platform.machine())
SDL_BINARY_PATTERNS: dict[str, list[str]] = \
    {"Windows": ["{}.dll"], "Darwin": ["lib{}.dylib", "{0}.framework/{0}", "{0}.framework/Versions/A/{0}"], "Linux": ["lib{}.so"]}

def SDL_PLATFORM_SPECIFIC(system: list[str] = None, arch: list[str] = None) -> bool:
    """Check if the current platform is inside the given platforms."""
    if int(os.environ.get("SDL_PLATFORM_AGNOSTIC", "0")) > 0: return True
    return (not system or SDL_SYSTEM in system) and (not arch or SDL_ARCH in arch)

__release__: bool = not int(os.environ.get("SDL_DEBUG", "0")) > 0
__frozen__: bool = getattr(sys, "frozen", False) or getattr(sys, "_MEIPASS", None) is not None

__doc_file__: str = os.path.join(os.path.dirname(__file__), "__doc__.py")
__doc_generator__: int = int(os.environ.get("SDL_DOC_GENERATOR", "0" if __frozen__ else "1")) > 0

__initialized__: bool = __name__.split(".")[0] in sys.modules if "." in __name__ else False
__module__: types.ModuleType = sys.modules[__name__.split(".")[0] if "." in __name__ else __name__]

class SDL_LOGGER:
    Info, Warning, Error = range(3)
    __level__: int = int(os.environ.get("SDL_LOG_LEVEL", "0"))
    __colors__: dict[int, str] = {Info: "\33[32m", Warning: "\33[35m", Error: "\33[31m"}
    __names__: dict[int, str] = {Info: "Info", Warning: "Warning", Error: "Error"}

    @classmethod
    def Log(cls, level: int, message: str | None = None, end: str | None = "\n") -> None:
        """Log a message at a specific level."""
        assert level in [cls.Error, cls.Warning, cls.Info], "Unknown log level."
        if cls.__level__ is None or level < cls.__level__: return

        if message is not None:
            sys.stdout.write(f"{cls.__colors__[level]}{cls.__names__[level]}: {message}\33[0m")

        if end is not None:
            sys.stdout.write(end)

        sys.stdout.flush()

def SDL_FIND_BINARIES(libraries: list[str]) -> list[str]:
    """Search for binaries in the system libraries."""
    libraries = libraries + [f"{library}d" for library in libraries]
    binaries = [f"./{file}" if SDL_SYSTEM in ["Windows"] and not ("/" in file or "\\" in file) else file for library in libraries if (file := ctypes.util.find_library(library))]

    if SDL_SYSTEM in ["Darwin"] and SDL_ARCH in ["ARM64"] and os.path.exists(path := "/opt/Homebrew/lib"):
        binaries += [file for library in libraries for pattern in SDL_BINARY_PATTERNS if os.path.exists(file := os.path.join(path, pattern.format(library)))]

    return binaries

def SDL_DOWNLOAD_BINARIES(path: str, system: str = SDL_SYSTEM, arch: str = SDL_ARCH) -> bool:
    """Download specific SDL3 binaries to the given path."""
    assert system.capitalize() in ["Darwin", "Linux", "Windows"], "Unknown system."
    assert arch.upper() in ["AMD64", "ARM64"], "Unknown architecture."
    if not os.path.exists(path): os.makedirs(path)

    headers = {"Accept": "application/vnd.github+json"}

    if len(os.environ.get("SDL_GITHUB_TOKEN", "")) != 0:
        headers["Authorization"] = f"Bearer {os.environ['SDL_GITHUB_TOKEN']}"

    try:
        for release in (_ := requests.get("https://api.github.com/repos/Aermoss/PySDL3-Build/releases", headers = headers).json()):
            if isinstance(release, str):
                SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to download binaries: {_[release]}")
                return False

            if release["draft"] or release["prerelease"]:
                continue

            for asset in release["assets"]:
                if asset["name"] != f"{system.capitalize()}-{arch.upper()}-{release['tag_name']}.zip":
                    continue

                with requests.get(asset["browser_download_url"], headers = headers, stream = True) as response:
                    if response.status_code != 200:
                        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to download binaries, status: {response.status_code} ({response.reason}).")
                        return False

                    size, current = int(response.headers.get("content-length", 0)), 0

                    with open(os.path.join(path, asset["name"]), "wb") as file:
                        for chunk in response.iter_content(chunk_size = 8192):
                            if chunk: file.write(chunk); current += len(chunk)
                            SDL_LOGGER.Log(SDL_LOGGER.Info, f"Downloading '{asset['browser_download_url']}'... {current / size * 100:.1f}% ({current / (1024 ** 2):.2f}/{size / (1024 ** 2):.2f} MB).", end = "\r")

                        SDL_LOGGER.Log(SDL_LOGGER.Info, end = "\n")

                    data = {
                        "arch": arch, "system": system, "target": f"v{__version__}",
                        "version": release["tag_name"], "url": asset["browser_download_url"],
                        "created": asset["created_at"], "updated": asset["updated_at"],
                        "uploader": asset["uploader"]["login"], "files": [],
                        "repair": True, "find": True
                    }

                    with zipfile.ZipFile(os.path.join(path, asset["name"]), "r") as ref:
                        for name in ref.namelist():
                            data["files"].append(f"./{name}")

                            if os.path.exists(os.path.join(path, name)):
                                os.remove(os.path.join(path, name))

                        ref.extractall(path)

                    with open(os.path.join(path, "metadata.json"), "w") as file:
                        json.dump(data, file, indent = 4)

                    os.remove(os.path.join(path, asset["name"]))
                    return True

            SDL_LOGGER.Log(SDL_LOGGER.Error, "Failed to download binaries, no binaries found.")
            return False

    except requests.RequestException as exc:
        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to download binaries: {exc}.")
        return False

if not __initialized__:
    if int(os.environ.get("SDL_CHECK_VERSION", "0" if __frozen__ else "1")) > 0:
        try:
            version = requests.get(f"https://pypi.org/pypi/PySDL3/json", timeout = 0.5).json()["info"]["version"]

            if packaging.version.parse(__version__) < packaging.version.parse(version):
                SDL_LOGGER.Log(SDL_LOGGER.Warning, f"You are using an older version of PySDL3 (current: {__version__}, latest: {version}).")

        except requests.RequestException:
            ...

    functions, binaryMap = {i: {} for i in SDL_BINARY_VAR_MAP_INV}, {}
    binaryData, missing = {"system": SDL_SYSTEM, "arch": SDL_ARCH, "files": []}, None
    binaryPath = os.environ.get("SDL_BINARY_PATH", os.path.join(os.path.dirname(__file__), "bin"))
    absPath = lambda path: path if os.path.isabs(path) else os.path.abspath(os.path.join(binaryPath, path))
    if not os.path.exists(binaryPath): os.makedirs(binaryPath)

    if int(os.environ.get("SDL_DISABLE_METADATA", "0")) > 0:
        for root, _, files in os.walk(binaryPath):
            for file in files:
                if (file.endswith(".dll") and SDL_SYSTEM in ["Windows"]) \
                or ((file.endswith(".dylib") or ".framework" in file) and SDL_SYSTEM in ["Darwin"]) \
                or (".so" in file and SDL_SYSTEM in ["Linux"]):
                    binaryData["files"].append(os.path.join(root, file))

    elif "metadata.json" in (files := os.listdir(binaryPath)):
        with open(os.path.join(binaryPath, "metadata.json"), "r") as file:
            missing, binaryData = True, json.load(file)
            binaryData["files"] = [absPath(i) for i in binaryData["files"]]

        if packaging.version.parse(__version__) > packaging.version.parse(binaryData.get("target", __version__)):
            SDL_LOGGER.Log(SDL_LOGGER.Warning, f"Incompatible target version detected: '{binaryData['target']}', current: 'v{__version__}'.")

        elif binaryData["system"].lower() != SDL_SYSTEM.lower() or binaryData["arch"].lower() != SDL_ARCH.lower():
            SDL_LOGGER.Log(SDL_LOGGER.Warning, f"Incompatible binary architecture and/or system detected: '{binaryData['system']} ({binaryData['arch']})'.")
            binaryData["repair"] = True

            for i in binaryData["files"]:
                if os.path.exists(i): os.remove(i)

        else:
            if missing := [i for i in binaryData["files"] if not os.path.exists(i)]:
                SDL_LOGGER.Log(SDL_LOGGER.Warning, f"Missing binary file(s) detected: '{', '.join(missing)}'.")

    else:
        SDL_LOGGER.Log(SDL_LOGGER.Warning, "No metadata detected.")
        missing = True

    if int(os.environ.get("SDL_DOWNLOAD_BINARIES", "1" if binaryData.get("repair", True) else "0")) > 0 and missing:
        if _ := SDL_DOWNLOAD_BINARIES(binaryPath, SDL_SYSTEM, SDL_ARCH):
            with open(os.path.join(binaryPath, "metadata.json"), "r") as file:
                binaryData, missing = json.load(file), None
                binaryData["files"] = [absPath(i) for i in binaryData["files"]]

    if int(os.environ.get("SDL_FIND_BINARIES", "1" if binaryData.get("find", missing is None) else "0")) > 0:
        binaryData["files"] += SDL_FIND_BINARIES(list(SDL_BINARY_VAR_MAP_INV.keys()))

    for binary in SDL_BINARY_VAR_MAP_INV:
        for path in binaryData["files"].copy():
            if (file := os.path.split(path)[1]).split(".")[0].endswith(binary) and os.path.exists(path):
                if (file.endswith(".dll") and SDL_SYSTEM in ["Windows"]) \
                or ((file.endswith(".dylib") or ".framework" in file) and SDL_SYSTEM in ["Darwin"]) \
                or (".so" in file and SDL_SYSTEM in ["Linux"]):
                    binaryMap[binary], functions[binary] = ctypes.CDLL(os.path.abspath(path)), {}
                    binaryData["files"].remove(path)
                    break

BaseType = typing.TypeVar("BaseType")
TargetType = typing.TypeVar("TargetType")

def SDL_ARRAY(*args: BaseType, **kwargs: TargetType) -> tuple[ctypes.Array[BaseType | TargetType], int]:
    """Create a ctypes array from the given arguments."""
    return ((kwargs.get("type", None) or args[0].__class__) * len(args))(*args), len(args)

def SDL_DEREFERENCE(value: typing.Any) -> typing.Any:
    """Dereference a ctypes pointer or object."""
    if isinstance(value, ctypes._Pointer): return value.contents
    elif hasattr(value, "_obj"): return value._obj
    else: return value

def SDL_CACHE_FUNC(func: abc.Callable[..., typing.Any]) -> abc.Callable[..., typing.Any]:
    """Simple function cache decorator."""
    cache = {}

    def __inner__(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        _hash = hash((args, tuple(frozenset(sorted(kwargs.items())))))
        if _hash not in cache: cache.update({_hash: func(*args, **kwargs)})
        return cache.get(_hash, None)

    return __inner__

@SDL_CACHE_FUNC
def SDL_GET_BINARY_NAME(binary: ctypes.CDLL) -> str | None:
    """Get the name of an SDL3 binary."""
    return {v: k for k, v in __module__.binaryMap.items()}.get(binary, None)

def SDL_GET_BINARY(name: str) -> ctypes.CDLL | None:
    """Get an SDL3 binary by its name."""
    return __module__.binaryMap.get(name, None)

def SDL_NOT_IMPLEMENTED(name: str) -> abc.Callable[..., None]:
    return lambda *args, **kwargs: \
        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Invoked an unimplemented function: '{name}'.")

class SDL_FUNC:
    @classmethod
    def __class_getitem__(cls, key: tuple[str, type, list[type], str]) -> typing.Any:
        """Create a new ctypes function definition."""

        if not (__frozen__ or __release__):
            assert isinstance(key, tuple), "Expected a tuple, got a single argument."
            assert len(key) == 4, "Expected a tuple with length 4."
            assert isinstance(key[0], str), "Expected a string as the first argument."
            assert isinstance(key[1], type) or key[1] is None, "Expected a type as the second argument."
            assert isinstance(key[2], list), "Expected a list as the third argument."
            assert ... not in key[2] or key[2].count(...) == 1, "Expected at most 1 '...' in the argument list."
            assert ... not in key[2] or key[2][-1] == ..., "Expected '...' at the end of the argument list."
            assert isinstance(key[3], str), "Expected a string as the fourth argument."
            assert key[3] in SDL_BINARY_VAR_MAP_INV, "Unknown binary."

        if binary := SDL_GET_BINARY(key[3]):
            func = getattr(binary, key[0], None)

            if not func and not int(os.environ.get("SDL_IGNORE_MISSING_FUNCTIONS", "1" if __frozen__ or __release__ else "0")) > 0:
                SDL_LOGGER.Log(SDL_LOGGER.Warning, f"Function '{key[0]}' not found in binary: '{SDL_BINARY_PATTERNS[SDL_SYSTEM][0].format(key[3])}'.")

        if not binary or not func: func = SDL_NOT_IMPLEMENTED(key[0])
        func.restype, func.binary = key[1], binary

        if ... in key[2]:
            def __inner__(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                for arg in args[len(__inner__.func.argtypes):]:
                    if isinstance(arg, int): __inner__.func.argtypes += [ctypes.c_int]
                    elif isinstance(arg, float): __inner__.func.argtypes += [ctypes.c_float]
                    elif isinstance(arg, bytes): __inner__.func.argtypes += [ctypes.c_char_p]
                    elif isinstance(arg, bool): __inner__.func.argtypes += [ctypes.c_bool]
                    else: __inner__.func.argtypes += [arg.__class__]

                value = __inner__.func(*args, **kwargs)
                __inner__.func.argtypes = key[2][:-1]
                return value

            func.argtypes, func.vararg = key[2][:-1], True
            func, __inner__.func = __inner__, func

        else:
            func.argtypes, func.vararg = key[2], False

        __module__.functions[key[3]][key[0]] = func
        return func

class SDL_POINTER:
    @classmethod
    def __class_getitem__(cls, key: type) -> type:
        """Create a ctypes pointer type from a ctypes type."""

        if not (__frozen__ or __release__):
            assert not isinstance(key, tuple), "Expected a single argument, got a tuple."
            assert isinstance(key, type), "Expected a type as the first argument."

        return ctypes.POINTER(key)

class SDL_CAST:
    @classmethod
    def __class_getitem__(cls, key: tuple[typing.Any, type]) -> typing.Any:
        """Cast a ctypes pointer to an another type."""

        if not (__frozen__ or __release__):
            assert isinstance(key, tuple), "Expected a tuple, got a single argument."
            assert len(key) == 2, "Expected a tuple with length 2."
            assert isinstance(key[0], ctypes._Pointer), "Expected a pointer as the first argument."
            assert isinstance(key[1], type), "Expected a type as the second argument."

        return ctypes.cast(key[0], key[1])

class SDL_TYPE:
    @classmethod
    def __class_getitem__(cls, key: tuple[str, type]) -> type:
        """Create a new type from a ctypes type."""

        if not (__frozen__ or __release__):
            assert isinstance(key, tuple), "Expected a tuple, got a single argument."
            assert len(key) == 2, "Expected a tuple with length 2."
            assert isinstance(key[0], str), "Expected a string as the first argument."
            assert isinstance(key[1], type), "Expected a type as the second argument."

        if hasattr(key[1], "contents"):
            return type(key[0], (ctypes._Pointer, ), {"_type_": key[1]._type_, "contents": key[1].contents})
        
        elif hasattr(key[1], "_fields_"):
            return type(key[0], (ctypes.Structure, ), {"_fields_": key[1]._fields_})

        else:
            return type(key[0], (ctypes._SimpleCData, ), {"_type_": key[1]._type_})

class SDL_FUNC_TYPE:
    @classmethod
    def __class_getitem__(cls, key: tuple[str, type, list[type]]) -> type:
        """Create a new ctypes function type."""

        if not (__frozen__ or __release__):
            assert isinstance(key, tuple), "Expected a tuple, got a single argument."
            assert len(key) == 3, "Expected a tuple with length 3."
            assert isinstance(key[0], str), "Expected a string as the first argument."
            assert isinstance(key[1], type) or key[1] is None, "Expected a type as the second argument."
            assert isinstance(key[2], list), "Expected a list as the third argument."

        _, ctypes._c_functype_cache = ctypes._c_functype_cache, {}
        value = ctypes.CFUNCTYPE(key[1], *key[2])
        value.__name__, ctypes._c_functype_cache = key[0], _
        return value

SDL_ENUM: typing.TypeAlias = SDL_TYPE["SDL_ENUM", ctypes.c_int]
SDL_VA_LIST: typing.TypeAlias = SDL_TYPE["SDL_VA_LIST", ctypes.c_char_p]

def SDL_PARSE_ARGUMENTS(argc: ctypes.c_int, argv: SDL_POINTER[ctypes.c_char_p]) -> list[str]:
    return [argv[i].decode("utf-8") for i in range(argc)]

def SDL_PROCESS_DESCRIPTION(description: str, url: str | None = None, rst: bool = False) -> str:
    """Process HTML description."""

    return re.sub(r"<a href=\"([a-zA-Z0-9_]+)\">([a-zA-Z0-9_]+)</a>", r"\1" if rst or not url else rf"[\2]({'/'.join(url.split('/')[:-1])}/\1)", description) \
        .replace("</", "<").replace("<em>", "" if rst else "*").replace("<code>", "`" if rst else "``").replace("<strong>", "" if rst else "**")

async def SDL_GET_LATEST_RELEASES() -> dict[str, str]:
    """Get latest releases of SDL3 modules from their official github repositories."""
    session, releases, tasks = aiohttp.ClientSession(), {}, []
    headers = {"Accept": "application/vnd.github+json"}

    if "SDL_GITHUB_TOKEN" in os.environ:
        headers["Authorization"] = f"Bearer {os.environ['SDL_GITHUB_TOKEN']}"

    for repo in SDL_REPOSITORIES:
        url = f"https://api.github.com/repos/libsdl-org/{repo}/releases"
        tasks.append(asyncio.create_task(session.get(url, headers = headers, ssl = False)))
        SDL_LOGGER.Log(SDL_LOGGER.Info, f"Sending a request to '{url}'.")

    responses = await asyncio.gather(*tasks)
    SDL_LOGGER.Log(SDL_LOGGER.Info, f"Response gathering completed ({len(responses)} responses).")

    for response, repo in zip(responses, SDL_REPOSITORIES):
        if response.status != 200:
            SDL_LOGGER.Log(SDL_LOGGER.Warning, f"Failed to get latest release of '{response.url}', skipping (status: {response.status}).")
            releases[repo] = None

        else:
            latestRelease = (None, None)

            for release in await response.json():
                score = packaging.version.parse(release["tag_name"].split("-")[1])

                if not latestRelease[0] or score > latestRelease[0]:
                    latestRelease = (score, release["tag_name"])

            releases[repo] = latestRelease[1]

    await session.close()
    return releases

async def SDL_GET_FUNC_DESCRIPTIONS(funcs: list[tuple[str, str]], rst: bool = False) -> tuple[list[str], list[list[str]]]:
    """Get descriptions, arguments and return types of SDL3 functions from the official SDL3 wiki."""
    session, tasks = aiohttp.ClientSession(), []

    for module, func in funcs:
        url = f"https://wiki.libsdl.org/{module}/{func}"
        SDL_LOGGER.Log(SDL_LOGGER.Info, f"Sending a request to '{url}'.")
        tasks.append(asyncio.create_task(session.get(url, ssl = False)))

    responses = await asyncio.gather(*tasks)
    SDL_LOGGER.Log(SDL_LOGGER.Info, f"Response gathering completed ({len(responses)} responses).")
    descriptions, arguments, returns = [], [], []

    for response in responses:
        if response.status != 200:
            SDL_LOGGER.Log(SDL_LOGGER.Warning, f"No such page: '{response.url}', skipping (status: {response.status}).")
            descriptions.append(None)
            arguments.append(None)
            returns.append(None)

        else:
            content = (await response.content.read()).decode()

            for a, b in zip(list(re.finditer(r"<p>", content)), list(re.finditer(r"</p>", content))):
                if content[a.end()] == "<": continue
                descriptions.append(SDL_PROCESS_DESCRIPTION(content[a.end():b.start()], str(response.url), rst))
                break

            for a, b in zip(list(re.finditer(r"<code([a-zA-Z0-9_=#\s\"\-]*)>", content)), list(re.finditer(r"</code>", content))):
                if "sourceCode c" not in content[a.start():a.end()]: continue
                text = re.sub(r"<(/)?([a-zA-Z0-9_=#\s\"\-]+)>", r"", content[a.end():b.start()]).split(";")[0]
                if "\n" in text: text = " ".join(text.split("\n"))
                returns.append(" ".join([i for i in text.split("(")[0].split(" ")[:-1] if i and i not in ["extern", "SDL_DECLSPEC"]]))
                text = text[text.index("(") + 1:text.index(")") + 1].replace(")", ",")[:-1]
                arguments.append({})

                for i in (text.split(", ") if "," in text else [text]) if text != "void" else []:
                    if i != "...": i = i.split(" ")
                    else: continue

                    arguments[-1][i[-1].replace("*", "")] = \
                        " ".join(i[:-1]) + "*" * i[-1].count("*")

                break

    await session.close()
    return descriptions, arguments, returns

async def SDL_GET_STRUCT_DESCRIPTIONS(structs: list[tuple[str, str]], rst: bool = False) -> tuple[list[str], list[list[str]]]:
    """Get descriptions and members of SDL3 structures from the official SDL3 wiki."""
    session, tasks = aiohttp.ClientSession(), []

    for module, struct in structs:
        url = f"https://wiki.libsdl.org/{module}/{struct}"
        SDL_LOGGER.Log(SDL_LOGGER.Info, f"Sending a request to '{url}'.")
        tasks.append(asyncio.create_task(session.get(url, ssl = False)))

    responses = await asyncio.gather(*tasks)
    SDL_LOGGER.Log(SDL_LOGGER.Info, f"Response gathering completed ({len(responses)} responses).")
    descriptions, members = [], []

    for response in responses:
        if response.status != 200:
            SDL_LOGGER.Log(SDL_LOGGER.Warning, f"No such page: '{response.url}', skipping (status: {response.status}).")
            descriptions.append(None)
            members.append(None)

        else:
            content = (await response.content.read()).decode()

            for a, b in zip(list(re.finditer(r"<p>", content)), list(re.finditer(r"</p>", content))):
                if content[a.end()] == "<": continue
                descriptions.append(SDL_PROCESS_DESCRIPTION(content[a.end():b.start()], str(response.url), rst))
                break

            for a, b in zip(list(re.finditer(r"<code([a-zA-Z0-9_=#\s\"\-]*)>", content)), list(re.finditer(r"</code>", content))):
                if "sourceCode c" not in content[a.start():a.end()]: continue
                text = re.sub(r"<(/)?([a-zA-Z0-9_=#\s\"\-]+)>", r"", content[a.end():b.start()])
                text, _ = re.sub(r"//.*", r"", re.sub(r"/\*[\s\S]*?\*/", r"", text)), text.split(" ")
                members.append({})

                for line in text.split("\n"):
                    if ";" not in line or "{" in line or "}" in line:
                        continue

                    line = line.split(";")[0].strip().replace(", ", ",")

                    if "SDLCALL" not in line:
                        count, line = line.count("*"), line.replace("*", "").split(" ")

                        for i in line[-1].split(",") if "," in line[-1] else [line[-1]]:
                            members[-1][i] = " ".join(line[:-1]) + "*" * count

                    else:
                        _ = line.replace("SDLCALL", "").split("(")
                        arguments, name = _[2].split(")")[0].strip(), _[1].split(")")[0].replace("*", "").strip()
                        members[-1][name] = {"return": _[0].strip(), "arguments": {}}

                        for i in arguments.split(",") if "," in arguments else [arguments]:
                            if i in ["...", "void"]: continue
                            count, i = i.count("*"), i.replace("*", "").split(" ")
                            members[-1][name]["arguments"][i[-1]] = " ".join(i[:-1]) + "*" * count

                break

    await session.close()
    return descriptions, members

def SDL_GET_MODULE_BY_NAME(name: str) -> str:
    """Get the module of an SDL3 function/structure by its name."""

    for prefix, module in sorted({"SDL": "SDL3", "IMG": "SDL3_image", "Mix": "SDL3_mixer", "TTF": "SDL3_ttf", "RTF": "SDL3_rtf", "NET": "SDL3_net", "SDL_ShaderCross": "SDL3_shadercross"}.items(), key = lambda x: -len(x[0])):
        if name.startswith(prefix): return module

def SDL_PYTHONIZE_TYPE(type: str, name: str | None = None, globals: dict[str, typing.Any] = {}) -> str:
    """Pythonize C/C++ types."""
    array = (_ if str.isnumeric(_ := name.split("[")[1].split("]")[0]) else eval(_, globals)) if name is not None and name.count("[") == 1 and name.index("]") - name.index("[") != 1 else None
    type, count = type.replace("*", ""), type.count("*") + (0 if name is None or array is not None else name.count("["))

    if " " in type:
        type = " ".join([i for i in type.split(" ") if i and i not in ["const", "struct", "union", "enum", "SDLCALL"]])

    if type in (sdlTypes := {
        "Sint8": "ctypes.c_int8", "Uint8": "ctypes.c_uint8", "Sint16": "ctypes.c_int16", "Uint16": "ctypes.c_uint16",
        "Sint32": "ctypes.c_int32", "Uint32": "ctypes.c_uint32", "Sint64": "ctypes.c_int64", "Uint64": "ctypes.c_uint64"
    }): type = sdlTypes[type]

    if count and type in ["void", "char", "wchar_t"]:
        type, count = f"ctypes.c_{type.replace('_t', '')}_p", count - 1

    if type in (cTypes := {
        "int": "ctypes.c_int", "bool": "ctypes.c_bool", "float": "ctypes.c_float", "double": "ctypes.c_double",
        "char": "ctypes.c_char", "short": "ctypes.c_short", "long long": "ctypes.c_longlong", "long": "ctypes.c_long",
        "size_t": "ctypes.c_size_t", "ssize_t": "ctypes.c_ssize_t", "intptr_t": "ctypes.c_intptr_t", "uintptr_t": "ctypes.c_uintptr_t",
        "int8_t": "ctypes.c_int8", "uint8_t": "ctypes.c_uint8", "int16_t": "ctypes.c_int16", "uint16_t": "ctypes.c_uint16",
        "int32_t": "ctypes.c_int32", "uint32_t": "ctypes.c_uint32", "int64_t": "ctypes.c_int64", "uint64_t": "ctypes.c_uint64",
        "unsigned short": "ctypes.c_ushort", "unsigned long long": "ctypes.c_ulonglong", "unsigned long": "ctypes.c_ulong",
        "unsigned char": "ctypes.c_ubyte", "unsigned int": "ctypes.c_uint", "wchar_t": "ctypes.c_wchar", "va_list": "SDL_VA_LIST"
    }): type = cTypes[type]

    if not count and type in ["void"]: type = "None"
    if count and type.startswith("ctypes."): type = type.split(".")[1]
    return f"{'LP_' * count}{type}{'_Array_%s' % array if array else ''}"

def SDL_GENERATE_DOCS(modules: list[str] = list(SDL_BINARY_VAR_MAP_INV.keys()), raw: types.ModuleType | None = None, rst: bool = False) -> str:
    """Generate type hints and docstring for SDL3 functions/structures using SDL3 wiki."""

    __index, (descriptions, arguments, returns) = -1, \
        asyncio.run(SDL_GET_FUNC_DESCRIPTIONS([(module, func) for module in modules for func in __module__.functions[module]], rst))

    for module in modules:
        for func in __module__.functions[module]:
            if (_ := __module__.functions[module][func]).__name__ == "__inner__": _ = _.func
            _.__doc__ = (descriptions[__index := __index + 1], arguments[__index], returns[__index])

    result = "" if rst else f"\"\"\"\n# This file is auto-generated, do not modify it.\n__meta__ = "
    if not rst: result += f"{{\"target\": \"v{__version__}\", \"system\": \"{SDL_SYSTEM}\"}}\n\"\"\"\n\n"
    result += f"from {'sdl3' if rst else ''}.SDL import *\n\n"
    result += f"from {'sdl3' if rst else ''}.__init__ import {'' if rst else 'raw, '}ctypes, typing, {'SDL_POINTER' if rst else ''}"
    result += "\n" if rst else f"\\\n{' ' * 4}SDL_POINTER, SDL_CLONE_METACLASS as SDL_CloneMeta\n\n"
    types, definitions = set(), ""

    def SDL_GET_FULL_NAME(type: type | None) -> str:
        if type is None: return "None"
        if type.__name__.startswith("LP_"): types.add(type.__name__)
        if type.__name__.startswith("c_"): return f"ctypes.{type.__name__}"
        else: return type.__name__

    if not rst and raw is not None:
        structs = [(SDL_GET_MODULE_BY_NAME(name), name) for name in dir(raw) if hasattr(getattr(raw, name), "_fields_") \
            and not name.startswith("_") and "_" in name and name not in ["SDL_GamepadBinding", "SDL_TLSID"]]

        for (module, name), description, members in zip(structs, *asyncio.run(SDL_GET_STRUCT_DESCRIPTIONS(structs))):
            if not description or not members:
                continue

            if len(members) != len(_fields_ := getattr(raw, name)._fields_):
                SDL_LOGGER.Log(SDL_LOGGER.Error, f"Member count mismatch for https://wiki.libsdl.org/{module}/{name} (expected: {len(members)}, got: {len(_fields_)}).")

            for field, member in zip(_fields_, members):
                if field[0] != (_ := member.split("[")[0] if "[" in member else member):
                    SDL_LOGGER.Log(SDL_LOGGER.Error, f"Member name mismatch for https://wiki.libsdl.org/{module}/{name} (expected: {_}, got: {field[0]}).")

                if isinstance(members[member], dict):
                    if (left := SDL_GET_FULL_NAME(field[1]._restype_)) != (right := SDL_PYTHONIZE_TYPE(members[member]["return"], globals = raw.__dict__)):
                        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Member return type mismatch for https://wiki.libsdl.org/{module}/{name} (member: {_}, expected: {right}, got: {left}).")

                    for left, (_name, right) in zip(field[1]._argtypes_, members[member]["arguments"].items()):
                        if (_left := SDL_GET_FULL_NAME(left)) != (_right := SDL_PYTHONIZE_TYPE(right, _name, globals = raw.__dict__)):
                            SDL_LOGGER.Log(SDL_LOGGER.Error, f"Member argument type mismatch for https://wiki.libsdl.org/{module}/{name} (member: {_}, argument: {_name}, expected: {_right}, got: {_left}).")

                else:
                    if (left := SDL_GET_FULL_NAME(field[1])) != (right := SDL_PYTHONIZE_TYPE(members[member], member, globals = raw.__dict__)):
                        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Member type mismatch for https://wiki.libsdl.org/{module}/{name} (member: {member}, expected: {right}, got: {left}).")

            result += f"class {name}(metaclass = SDL_CloneMeta):\n{' ' * 4}\"\"\"{description}{'' if description.endswith('.') else '.'}\"\"\"\n\n"
            # result += f"{' ' * 4}def __init__({', '.join(["self"] + [f'{_[0]}: {SDL_GET_FULL_NAME(_[1])}' for _ in _fields_])}) -> None:\n"
            # result += f"{' ' * 8}super({name}, self).__init__()\n\n{' ' * 8}{', '.join([f'self.{_[0]}' for _ in _fields_])} = {', '.join([_[0] for _ in _fields_])}\n\n"

    for index, module in enumerate(modules):
        if len(__module__.functions[module]) == 0: continue
        if not rst: definitions += f"# {SDL_BINARY_PATTERNS[SDL_SYSTEM][0].format(module)} implementation.\n\n"

        for _index, func in enumerate(__module__.functions[module]):
            if (_ := __module__.functions[module][func]).__name__ == "__inner__": _ = _.func
            vararg, restype, argtypes, (description, arguments, _return) = _.vararg, _.restype, _.argtypes, _.__doc__

            if arguments is not None and len(arguments) != len(argtypes):
                SDL_LOGGER.Log(SDL_LOGGER.Error, f"Argument count mismatch for 'https://wiki.libsdl.org/{module}/{func}' (expected: {len(arguments)}, got: {len(argtypes)}).")

            if arguments is None:
                arguments = [f"_{i}" for i in range(len(argtypes))]

            else:
                for __index, i in enumerate(arguments):
                    if (left := SDL_PYTHONIZE_TYPE(arguments[i], i)) != (right := SDL_GET_FULL_NAME(argtypes[__index])):
                        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Argument type mismatch for 'https://wiki.libsdl.org/{module}/{func}' (argument: {i}, expected: {left}, got: {right}).")

                arguments = [f"{'_' if i in keyword.kwlist else ''}{i.replace('[', '').replace(']', '')}" for i in arguments]

            if _return is not None and (left := SDL_PYTHONIZE_TYPE(_return)) != (right := SDL_GET_FULL_NAME(restype)):
                SDL_LOGGER.Log(SDL_LOGGER.Error, f"Return type mismatch for 'https://wiki.libsdl.org/{module}/{func}' (expected: {left}, got: {right}).")

            definitions += f"def {func}({', '.join([f'{arg}: {SDL_GET_FULL_NAME(type)}' for arg, type in zip(arguments, argtypes)] + (['*args: typing.Any'] if vararg else []))}) -> {SDL_GET_FULL_NAME(restype)}:\n"

            if not rst or description is not None: definitions += f"{' ' * 4}\"\"\"\n"
            if description is not None: definitions += f"    {description}\n"
            if not rst: definitions += f"\n{' ' * 4}https://wiki.libsdl.org/{module}/{func}\n"
            if not rst or description is not None: definitions += f"{' ' * 4}\"\"\""
            if not rst: definitions += f"\n{' ' * 4}return raw.{func}({', '.join(arguments + (['*args'] if vararg else []))})"
            if rst and description is None: definitions += f"\n{' ' * 4}..."

            if _index != len(__module__.functions[module]) - 1:
                definitions += "\n\n"

        if index != len(modules) - 1 and len(__module__.functions[modules[index + 1]]) != 0:
            definitions += "\n\n"

    for i in types:
        count, func = i.count("LP_"), i.replace("LP_", "")
        result += f"{i}: typing.TypeAlias = {'SDL_POINTER[' * count}{'ctypes.' if func.startswith('c_') else ''}{func}{']' * count}\n"

    return f"{result}\n{definitions}"

def SDL_GET_OR_GENERATE_DOCS(*args: typing.Any, **kwargs: typing.Any) -> bytes:
    """Get type hints and docstrings for SDL3 functions/structures from github or by generating them."""

    try:
        headers = {"Accept": "application/vnd.github+json"}

        if "SDL_GITHUB_TOKEN" in os.environ:
            headers["Authorization"] = f"Bearer {os.environ['SDL_GITHUB_TOKEN']}"

        for release in (_ := requests.get(f"https://api.github.com/repos/Aermoss/PySDL3/releases", headers = headers).json()):
            if isinstance(release, str):
                SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to get docs from github: {_[release]}")
                break

            if release["tag_name"] != f"v{__version__}":
                continue

            for asset in release["assets"]:
                if asset["name"] != f"{SDL_SYSTEM}-Docs.py":
                    continue

                with requests.get(asset["browser_download_url"], headers = headers, stream = True) as response:
                    if response.status_code != 200:
                        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to get docs from github, status: {response.status_code} ({response.reason}).")
                        break

                    return bytearray().join([chunk for chunk in response.iter_content(chunk_size = 8192) if chunk])

    except requests.RequestException as exc:
        SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to get docs from github: {exc}.")

    SDL_LOGGER.Log(SDL_LOGGER.Warning, "Generating docs... this may take a while.")
    return SDL_GENERATE_DOCS(*args, **kwargs).encode("utf-8")

if not __initialized__ and int(os.environ.get("SDL_CTYPES_ALIAS_FIX", "0" if __frozen__ or __release__ else "1")) > 0:
    for i in dir(ctypes):
        if i.startswith("c_") and getattr(ctypes, i).__name__ != i and hasattr(getattr(ctypes, i), "_type_"):
            setattr(ctypes, i, SDL_TYPE[i, getattr(ctypes, i)])

from sdl3.SDL import *

if __doc_generator__:
    import sdl3.SDL as raw

    class SDL_CLONE_METACLASS:
        """Simple clone metaclass."""

        def __new__(cls, name: str, _: typing.Any, attributes: dict[str, typing.Any]) -> typing.Any:
            _class = getattr(raw, name)
            
            for key, value in attributes.items():
                if key in ["__init__"]:
                    setattr(_class, key, value)

            return _class

SDL_VERSIONNUM_STRING: abc.Callable[[int], str] = lambda num: str(None).lower() if not num else \
    f"{SDL_VERSIONNUM_MAJOR(num)}.{SDL_VERSIONNUM_MINOR(num)}.{SDL_VERSIONNUM_MICRO(num)}"

if not __initialized__:
    if int(os.environ.get("SDL_CHECK_BINARY_VERSION", "1")) > 0:
        for binary, left, right in zip(SDL_BINARY_VAR_MAP.values(), [SDL_GetVersion, IMG_Version, Mix_Version, TTF_Version, RTF_Version, NET_GetVersion], [SDL_VERSION, SDL_IMAGE_VERSION, SDL_MIXER_VERSION, SDL_TTF_VERSION, SDL_RTF_VERSION, SDL_NET_VERSION]):
            if binary in binaryMap and (_ := left()) != right: SDL_LOGGER.Log(SDL_LOGGER.Warning, f"Version mismatch with binary: '{SDL_BINARY_PATTERNS[SDL_SYSTEM][0].format(binary)}' (expected: {SDL_VERSIONNUM_STRING(right)}, got: {SDL_VERSIONNUM_STRING(_)}).")

    def SDL_TRY_WRITE_DOCS() -> bool | None:
        try:
            with open(__doc_file__, "wb") as file:
                file.write(SDL_GET_OR_GENERATE_DOCS(raw = raw))

            return True

        except OSError as exc:
            SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to write docs: {exc}.")
    
    if __doc_generator__ and (os.path.exists(__doc_file__) or SDL_TRY_WRITE_DOCS()):
        try:
            from .__doc__ import *
            exec(getattr(__doc__, "__doc__"), data := {})

        except (SyntaxError, NameError, TypeError) as exc:
            SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to read docs: {exc}.")
            data = None

        if os.path.exists(__doc_file__) and (not data or "__meta__" not in data or data["__meta__"]["target"] != f"v{__version__}" or data["__meta__"]["system"] != SDL_SYSTEM) and SDL_TRY_WRITE_DOCS():
            if "sdl3.__doc__" in sys.modules: del sys.modules["sdl3.__doc__"]
            SDL_LOGGER.Log(SDL_LOGGER.Warning, "Reloading module: 'sdl3.__doc__'...")
            from .__doc__ import *

    else:
        try:
            if not (__frozen__ or __release__):
                if os.path.exists(__doc_file__): os.remove(__doc_file__)

        except PermissionError as exc:
            SDL_LOGGER.Log(SDL_LOGGER.Error, f"Failed to remove docs: {exc}.")


from .__init__ import ctypes, typing, abc, SDL_POINTER, \
    SDL_FUNC_TYPE, SDL_FUNC, SDL_TYPE, SDL_BINARY, SDL_ENUM

from .SDL_video import SDL_Window
from .SDL_properties import SDL_PropertiesID

class SDL_DialogFileFilter(ctypes.Structure):
    _fields_ = [
        ("name", ctypes.c_char_p),
        ("pattern", ctypes.c_char_p)
    ]

SDL_DialogFileCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_DialogFileCallback", None, [ctypes.c_void_p, SDL_POINTER[ctypes.c_char_p], ctypes.c_int]]

SDL_ShowOpenFileDialog: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ShowOpenFileDialog", None, [SDL_DialogFileCallback, ctypes.c_void_p, SDL_POINTER[SDL_Window], SDL_POINTER[SDL_DialogFileFilter], ctypes.c_int, ctypes.c_char_p, ctypes.c_bool], SDL_BINARY]
SDL_ShowSaveFileDialog: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ShowSaveFileDialog", None, [SDL_DialogFileCallback, ctypes.c_void_p, SDL_POINTER[SDL_Window], SDL_POINTER[SDL_DialogFileFilter], ctypes.c_int, ctypes.c_char_p], SDL_BINARY]
SDL_ShowOpenFolderDialog: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ShowOpenFolderDialog", None, [SDL_DialogFileCallback, ctypes.c_void_p, SDL_POINTER[SDL_Window], ctypes.c_char_p, ctypes.c_bool], SDL_BINARY]

SDL_FileDialogType: typing.TypeAlias = SDL_TYPE["SDL_FileDialogType", SDL_ENUM]

SDL_FILEDIALOG_OPENFILE, SDL_FILEDIALOG_SAVEFILE, SDL_FILEDIALOG_OPENFOLDER = range(3)

SDL_ShowFileDialogWithProperties: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ShowFileDialogWithProperties", None, [SDL_FileDialogType, SDL_DialogFileCallback, ctypes.c_void_p, SDL_PropertiesID], SDL_BINARY]

SDL_PROP_FILE_DIALOG_FILTERS_POINTER: bytes = "SDL.filedialog.filters".encode()
SDL_PROP_FILE_DIALOG_NFILTERS_NUMBER: bytes = "SDL.filedialog.nfilters".encode()
SDL_PROP_FILE_DIALOG_WINDOW_POINTER: bytes = "SDL.filedialog.window".encode()
SDL_PROP_FILE_DIALOG_LOCATION_STRING: bytes = "SDL.filedialog.location".encode()
SDL_PROP_FILE_DIALOG_MANY_BOOLEAN: bytes = "SDL.filedialog.many".encode()
SDL_PROP_FILE_DIALOG_TITLE_STRING: bytes = "SDL.filedialog.title".encode()
SDL_PROP_FILE_DIALOG_ACCEPT_STRING: bytes = "SDL.filedialog.accept".encode()
SDL_PROP_FILE_DIALOG_CANCEL_STRING: bytes = "SDL.filedialog.cancel".encode()